##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Pack/Undo implementations.
"""
from __future__ import absolute_import
from __future__ import print_function

import logging
import time

from ZODB.POSException import UndoError
from ZODB.utils import u64
from zope.interface import implementer

from .._compat import metricmethod
from .._compat import OidList
from .._util import byte_display
from .._util import get_memory_usage

from ..treemark import TreeMarker

from .schema import Schema
from .connections import LoadConnection
from .connections import PrePackConnection
from .interfaces import IPackUndo
from ._util import DatabaseHelpersMixin
from .sql import it

# pylint:disable=too-many-lines,unused-argument

logger = logging.getLogger(__name__)

class PackUndo(DatabaseHelpersMixin):
    """Abstract base class for pack/undo"""

    _choose_pack_transaction_query = None


    _lock_for_share = 'FOR SHARE'
    _lock_for_update = 'FOR UPDATE'

    driver = None
    connmanager = None
    runner = None
    locker = None
    options = None

    cursor_arraysize = 10000

    def __init__(self, database_driver, connmanager, runner, locker, options):
        self.driver = database_driver
        self.connmanager = connmanager
        self.runner = runner
        self.locker = locker
        self.options = options

    def with_options(self, options):
        """
        Return a new instance that will use the given options, instead
        of the options originally constructed.
        """
        if options == self.options:
            # If the options haven't changed, return ourself. This is
            # for tests that make changes to the structure of this
            # object not captured in the constructor or options.
            # (checkPackWhileReferringObjectChanges)
            return self
        return self.__class__(self.driver, self.connmanager, self.runner, self.locker, options)

    def choose_pack_transaction(self, pack_point):
        """Return the transaction before or at the specified pack time.

        Returns None if there is nothing to pack.
        """
        conn, cursor = self.connmanager.open()
        try:
            self._choose_pack_transaction_query.execute(cursor, {'tid': pack_point})
            rows = cursor.fetchall()
            if not rows:
                # Nothing needs to be packed.
                return None
            return rows[0][0]
        finally:
            self.connmanager.close(conn, cursor)

    def _make_ss_load_cursor(self, load_connection):
        ss_load_cursor = load_connection.server_side_cursor()
        ss_load_cursor.arraysize = self.cursor_arraysize
        try:
            ss_load_cursor.itersize = self.cursor_arraysize
        except AttributeError:
            pass
        return ss_load_cursor

    # Subclasses (notably Oracle) can define this to provide hints
    # that affect graph traversal.
    #
    # We cannot include the hints Oracle wants as standard; /*+ ... */
    # is also the syntax for a MySQL 5.7 optimizer hint, but FULL(...)
    # isn't valid syntax, so it produces a warning (and some
    # frameworks/drivers want to treat warnings as errors, or print
    # them).
    #
    # The alternate comment syntax for Oracle hints, --+ ..., isn't a
    # valid MySQL comment (MySQL requires whitespace after --) and raises
    # a syntax error.
    #
    # PostgreSQL doesn't have hints, so this is a no-op there.
    _traverse_graph_optimizer_hint = ''

    def _traverse_graph(self, load_connection, store_connection):
        """
        Visit the entire object graph to find out what should be
        kept.

        Sets the pack_object.keep flags.

        Must not read from the ``object_state`` table or any other table that
        could be inconsistent with the original snapshot view of references
        established by :meth:`pre_pack`.

        *cursor* is a writable store connection cursor.
        """
        logger.info("pre_pack: downloading pack_object and object_ref.")
        # Ensure we're up-to-date and can view the data in pack_object.
        # Note that we don't just use restart() here: if we haven't actually
        # opened the cursor yet, restart() won't do anything. But on MySQL,
        # (because we TRUNCATE'd the pack_object table?) if we don't actually
        # rollback, we get
        #   OperationalError: 1412, 'Table definition has changed, please retry transaction'
        load_connection.rollback_quietly()

        marker = TreeMarker()

        # Download the graph of object references into the TreeMarker.
        # TODO: We can probably do much or most of this in SQL, at least
        # in recent databases that support recursive WITH queries?

        # XXX: In history-free mode, ``pack_object`` contains exactly
        # the set of OIDs that are present in ``object_state`` and I
        # *think* that ``pack_object.keep_tid`` is always going to be equal to
        # ``object_ref.tid``. We may get better behaviour if we join
        # against that table here
        with load_connection.server_side_cursor() as ss_load_cursor:
            stmt = """
            SELECT {}
                zoid, object_ref.to_zoid
            FROM object_ref
                INNER JOIN pack_object USING (zoid)
            WHERE object_ref.tid >= pack_object.keep_tid
            ORDER BY zoid, object_ref.to_zoid
            """.format(self._traverse_graph_optimizer_hint)
            ss_load_cursor.execute(stmt)

            while True:
                rows = ss_load_cursor.fetchmany(self.cursor_arraysize)
                if not rows:
                    break
                marker.add_refs(rows)

        # Use the TreeMarker to find all reachable objects, starting
        # with the ones that are known reachable. These are the roots:
        #
        # - ZOID 0 which is explicitly marked as such
        #
        # - In history preserving databases where we are not doing GC,
        #   this includes all objects (except those explicitly
        #   deleted) --- but we don't actually call this method for
        #   the HP-no-gc case.
        #
        # - In history preserving *with* gc, this is all objects that
        #   have been modified after the pack time or are referenced
        #   from objects that have been modified after the pack time.
        #
        # - In history free *with* gc, this is all objects that have
        #   been modified after the pack time.
        # XXX: It seems like a lot of what TreeMarker does could actually
        # be done in the database, especially if we have support for
        # recursive common table expressions; if we don't, we can still do more of it
        # in the DB, it will just take more queries and some temp tables.
        logger.info("pre_pack: traversing the object graph "
                    "to find reachable objects.")
        with load_connection.server_side_cursor() as ss_load_cursor:
            stmt = """
            SELECT zoid
            FROM pack_object
            WHERE keep = %(TRUE)s
            """
            self.runner.run_script_stmt(ss_load_cursor, stmt)
            while True:
                rows = ss_load_cursor.fetchmany(self.cursor_arraysize)
                if not rows:
                    break
                marker.mark(oid for (oid,) in rows)

        marker.free_refs()

        # Upload the TreeMarker results to the database.
        # TODO: It probably makes more sense to mark *unreachable* objects?
        # There should generally be fewer of them than reachable objects
        # if the database is regularly GC'd.
        logger.info(
            "pre_pack: marking objects reachable: %d",
            marker.reachable_count)

        store_cursor = store_connection.cursor
        batch = []
        def upload_batch():
            # This would be easily done in parallel.
            # Marking 30 MM objects, even in batches, takes at least
            # 30 minutes.
            # XXX: No, this is very wrong. Use the RowBatcher
            # Alternately, flush to a table and then join
            # against that at the end.
            # TODO: If we're batching, we need to log progress
            oids_str = ','.join(str(oid) for oid in batch)
            del batch[:]
            stmt = """
            UPDATE pack_object
            SET keep = %%(TRUE)s,
                visited = %%(TRUE)s
            WHERE zoid IN (%s)
            """ % oids_str
            self.runner.run_script_stmt(store_cursor, stmt)

        batch_append = batch.append
        for oid in marker.reachable:
            batch_append(oid)
            if len(batch) >= 1000:
                upload_batch()
        if batch:
            upload_batch()

    # The only things to worry about are object_state and blob_chuck
    # and, in history-preserving, transaction. blob chunks are deleted
    # automatically by a foreign key; transaction we'll handle with a
    # pack. (We don't do anything with current_object; a state of NULL
    # represents a deleted object; it shouldn't be reachable anyway
    # and will be packed away next time we pack (without GC))

    # We shouldn't *have* to verify the oldserial in the delete statement,
    # because our only consumer is zc.zodbdgc which only calls us for
    # unreachable objects, so they shouldn't be modified and get a new
    # TID. But it's safer to do so.
    _script_delete_object = None

    def deleteObject(self, cursor, oid, oldserial):
        params = {'oid': u64(oid), 'tid': u64(oldserial)}
        self.runner.run_script_stmt(
            cursor,
            self._script_delete_object,
            params)
        return cursor.rowcount

    def on_filling_object_refs_added(self, oids=None, tids=None):
        """Test injection point for packing."""

@implementer(IPackUndo)
class HistoryPreservingPackUndo(PackUndo):
    """
    History-preserving pack/undo.
    """

    keep_history = True

    _choose_pack_transaction_query = Schema.transaction.select(
        it.c.tid
    ).where(
        it.c.tid > 0
    ).and_(
        it.c.tid <= it.bindparam('tid')
    ).and_(
        it.c.packed == False  # pylint:disable=singleton-comparison
    ).order_by(
        it.c.tid, 'DESC'
    ).limit(1)

    _script_create_temp_pack_visit = """
    CREATE TEMPORARY TABLE temp_pack_visit (
        zoid BIGINT NOT NULL PRIMARY KEY,
        keep_tid BIGINT NOT NULL
    );
    CREATE INDEX temp_pack_keep_tid ON temp_pack_visit (keep_tid)
    """

    _script_create_temp_undo = """
    CREATE TEMPORARY TABLE temp_undo (
        zoid BIGINT NOT NULL,
        prev_tid BIGINT NOT NULL
    );
    CREATE UNIQUE INDEX temp_undo_zoid ON temp_undo (zoid)
    """

    _script_reset_temp_undo = "DROP TABLE temp_undo"

    _choose_pack_tid_query = Schema.pack_object.select(
        it.c.keep_tid
    ).order_by(
        it.c.keep_tid, 'DESC'
    ).limit(1)

    _transaction_has_data_query = Schema.object_state.select(
        1
    ).where(
        it.c.tid == it.bindparam('tid')
    ).limit(1)


    _script_pack_current_object = """
    DELETE FROM current_object
    WHERE tid = %(tid)s
    AND zoid in (
        SELECT pack_state.zoid
        FROM pack_state
        WHERE pack_state.tid = %(tid)s
        %(INNER_ORDER_BY)s
    )
    """

    _script_pack_object_state = """
    DELETE FROM object_state
    WHERE tid = %(tid)s
    AND zoid in (
        SELECT pack_state.zoid
        FROM pack_state
        WHERE pack_state.tid = %(tid)s
        %(INNER_ORDER_BY)s
    )
    """

    _script_pack_object_ref = """
    DELETE FROM object_refs_added
    WHERE tid IN (
        SELECT tid
        FROM "transaction"
        WHERE is_empty = %(TRUE)s
    );

    DELETE FROM object_ref
    WHERE tid IN (
        SELECT tid
        FROM "transaction"
        WHERE is_empty = %(TRUE)s
    );
        """

    # Previously we used `= ANY(ARRAY(...))`, as was once recommended,
    # (See http://www.postgres.cz/index.php/PostgreSQL_SQL_Tricks#Fast_first_n_rows_removing)
    # but that is no longer recommended or expected to be faster.
    # Also, it was postgres specific. Now we use a more standard syntax,
    # that lets us preserve order (in case that matters).
    # pylint:disable=singleton-comparison
    _delete_empty_transactions_batch_query = Schema.transaction.delete(
    ).where(
        it.c.tid.in_(Schema.transaction.select(
            it.c.tid
        ).where(
            it.c.packed == True
        ).and_(
            it.c.is_empty == True
        ).order_by(
            it.c.tid
        ).limit(1000))
    )

    _script_delete_object = """
    UPDATE object_state
    SET state = NULL,
        state_size = 0,
        md5 = ''
    WHERE zoid = %(oid)s
    AND    tid = %(tid)s
    """

    _is_packed_tx_query = Schema.transaction.select(
        1
    ).where(
        Schema.transaction.c.tid == Schema.transaction.bindparam('undo_tid')
    ).and_(
        Schema.transaction.c.packed == False # pylint:disable=singleton-comparison
    )

    _is_root_creation_tx_query = Schema.object_state.select(
        1
    ).where(
        Schema.object_state.c.tid == Schema.object_state.bindparam('undo_tid')
    ).and_(
        Schema.object_state.c.zoid == 0
    ).and_(
        Schema.object_state.c.prev_tid == 0
    )

    @metricmethod
    def verify_undoable(self, cursor, undo_tid):
        """Raise UndoError if it is not safe to undo the specified txn."""
        self._is_packed_tx_query.execute(cursor, {'undo_tid': undo_tid})
        if not cursor.fetchall():
            raise UndoError("Transaction not found or packed")

        # Rule: we can undo an object if the object's state in the
        # transaction to undo matches the object's current state. If
        # any object in the transaction does not fit that rule, refuse
        # to undo. In theory this means arbitrary transactions can be
        # undone (because we actually match the MD5 of the state); in practice it
        # means that it must be the most recent transaction those
        # objects were involved in.

        # (Note that this prevents conflict-resolving undo as described
        # by ZODB.tests.ConflictResolution.ConflictResolvingTransUndoStorage.
        # Do people need that? If so, we can probably support it, but it
        # will require additional code.)
        stmt = """
        SELECT prev_os.zoid, current_object.tid
        FROM object_state prev_os
        INNER JOIN object_state cur_os
            ON (prev_os.zoid = cur_os.zoid)
        INNER JOIN current_object
            ON (cur_os.zoid = current_object.zoid
                AND cur_os.tid = current_object.tid)
        WHERE prev_os.tid = %s
            AND cur_os.md5 != prev_os.md5
        ORDER BY prev_os.zoid
        """
        self.runner.run_script_stmt(cursor, stmt, (undo_tid,))
        if cursor.fetchmany():
            raise UndoError(
                "Some data were modified by a later transaction")

        # Rule: don't allow the creation of the root object to
        # be undone.  It's hard to get it back.
        self._is_root_creation_tx_query.execute(cursor, {'undo_tid': undo_tid})
        if cursor.fetchall():
            raise UndoError("Can't undo the creation of the root object")


    @metricmethod
    def undo(self, cursor, undo_tid, self_tid):
        """Undo a transaction.

        Parameters: "undo_tid", the integer tid of the transaction to undo,
        and "self_tid", the integer tid of the current transaction.

        Returns the states copied forward by the undo operation as a
        list of (oid, old_tid).
        """
        stmt = self._script_create_temp_undo
        if stmt:
            self.runner.run_script(cursor, stmt)

        stmt = """
        DELETE FROM temp_undo;

        -- Put into temp_undo the list of objects to be undone and
        -- the tid of the transaction that has the undone state.
        INSERT INTO temp_undo (zoid, prev_tid)
        SELECT zoid, prev_tid
        FROM object_state
        WHERE tid = %(undo_tid)s
        ORDER BY zoid;

        -- Override previous undo operations within this transaction
        -- by resetting the current_object pointer and deleting
        -- copied states from object_state.
        UPDATE current_object
        SET tid = (
            SELECT prev_tid
            FROM object_state
            WHERE zoid = current_object.zoid
              AND tid = %(self_tid)s
        )
        WHERE zoid IN (SELECT zoid FROM temp_undo %(INNER_ORDER_BY)s)
        AND tid = %(self_tid)s;

        DELETE FROM object_state
        WHERE zoid IN (SELECT zoid FROM temp_undo %(INNER_ORDER_BY)s)
            AND tid = %(self_tid)s;

        -- Copy old states forward.
        INSERT INTO object_state (zoid, tid, prev_tid, md5, state_size, state)
        SELECT temp_undo.zoid, %(self_tid)s, current_object.tid,
            md5, COALESCE(state_size, 0), state
        FROM temp_undo
        INNER JOIN current_object ON (temp_undo.zoid = current_object.zoid)
        LEFT OUTER JOIN object_state
            ON (object_state.zoid = temp_undo.zoid
                AND object_state.tid = temp_undo.prev_tid)
        ORDER BY current_object.zoid;

        -- Copy old blob chunks forward.
        INSERT INTO blob_chunk (zoid, tid, chunk_num, chunk)
        SELECT temp_undo.zoid, %(self_tid)s, chunk_num, chunk
        FROM temp_undo
            JOIN blob_chunk
                ON (blob_chunk.zoid = temp_undo.zoid
                    AND blob_chunk.tid = temp_undo.prev_tid);

        -- List the copied states.
        SELECT zoid, prev_tid
        FROM temp_undo;
        """
        self.runner.run_script(cursor, stmt,
                               {'undo_tid': undo_tid, 'self_tid': self_tid})
        res = list(cursor)

        stmt = self._script_reset_temp_undo
        if stmt:
            self.runner.run_script(cursor, stmt)

        return res

    def fill_object_refs(self, load_connection, store_connection, get_references):
        """Update the object_refs table by analyzing new transactions."""
        with self._make_ss_load_cursor(load_connection) as ss_load_cursor:
            stmt = """
            SELECT tx.tid
            FROM "transaction" tx
            LEFT OUTER JOIN object_refs_added
                ON (tx.tid = object_refs_added.tid)
            WHERE object_refs_added.tid IS NULL
            ORDER BY tx.tid
            """
            self.runner.run_script_stmt(ss_load_cursor, stmt)
            tids = OidList((tid for (tid,) in ss_load_cursor))
        log_at = time.time() + 60
        tid_count = len(tids)
        txns_done = 0
        self.on_filling_object_refs_added(tids=tids)
        logger.info(
            "pre_pack: analyzing references from objects in %d new "
            "transaction(s)", tid_count)
        for tid in tids:
            self._add_refs_for_tid(load_connection, store_connection, tid, get_references)
            txns_done += 1
            now = time.time()
            if now >= log_at:
                # save the work done so far
                store_connection.commit()
                log_at = now + 60
                logger.info(
                    "pre_pack: transactions analyzed: %d/%d",
                    txns_done, tid_count)
        store_connection.commit()
        logger.info("pre_pack: transactions analyzed: %d/%d", txns_done, tid_count)

    _get_objects_in_transaction_query = Schema.object_state.select(
        it.c.zoid,
        it.c.state
    ).where(
        it.c.tid == it.bindparam('tid')
    ).order_by(it.c.zoid).prepared()

    _delete_refs_for_transaction_query = Schema.object_ref.delete(
    ).where(
        it.c.tid == it.bindparam('tid')
    ).prepared()

    _insert_object_refs_added_for_transaction_query = Schema.object_refs_added.insert(
        it.c.tid
    ).prepared()

    def _add_refs_for_tid(self, load_connection, store_connection, tid, get_references):
        """Fill object_refs with all states for a transaction.

        Returns the number of references added.
        """
        logger.debug("pre_pack: transaction %d: computing references ", tid)
        from_count = 0
        self._get_objects_in_transaction_query.execute(load_connection.cursor,
                                                       {'tid': tid})

        add_rows = []  # [(from_oid, tid, to_oid)]
        for from_oid, state in load_connection.cursor:
            state = self.driver.binary_column_as_state_type(state)
            if state:
                assert isinstance(state, self.driver.state_types), type(state)
                from_count += 1
                try:
                    to_oids = get_references(state)
                except:
                    logger.exception(
                        "pre_pack: can't unpickle "
                        "object %d in transaction %d; state length = %d",
                        from_oid, tid, len(state))
                    raise
                for to_oid in to_oids:
                    add_rows.append((from_oid, tid, to_oid))

        # A previous pre-pack may have been interrupted.  Delete rows
        # from the interrupted attempt.
        self._delete_refs_for_transaction_query.execute(store_connection.cursor,
                                                        {'tid': tid})

        # Add the new references.
        # TODO: Use RowBatcher?
        stmt = """
        INSERT INTO object_ref (zoid, tid, to_zoid)
        VALUES (%s, %s, %s)
        """
        self.runner.run_many(store_connection.cursor, stmt, add_rows)

        # The references have been computed for this transaction
        self._insert_object_refs_added_for_transaction_query.execute(store_connection.cursor,
                                                                     (tid,))

        to_count = len(add_rows)
        logger.debug("pre_pack: transaction %d: has %d reference(s) "
                     "from %d object(s)", tid, to_count, from_count)
        return to_count

    @metricmethod
    def pre_pack(self, pack_tid, get_references):
        """Decide what to pack.

        pack_tid specifies the most recent transaction to pack.

        get_references is a function that accepts a pickled state and
        returns a set of OIDs that state refers to.

        The self.options.pack_gc flag indicates whether
        to run garbage collection.
        If pack_gc is false, at least one revision of every object is kept,
        even if nothing refers to it.  Packing with pack_gc disabled can be
        much faster.
        """
        load_connection = LoadConnection(self.connmanager)
        store_connection = PrePackConnection(self.connmanager)
        try:
            # The pre-pack functions are responsible for managing
            # their own commits; when they return, the transaction
            # should be committed.
            #
            # ``pack_object`` should be populated,
            # essentially with the distinct list of all objects and their
            # maximum (newest) transaction ids.
            if self.options.pack_gc:
                logger.info("pre_pack: start with gc enabled")
                self._pre_pack_with_gc(
                    load_connection, store_connection, pack_tid, get_references)
            else:
                logger.info("pre_pack: start without gc")
                self._pre_pack_without_gc(
                    load_connection, store_connection, pack_tid)

            logger.info("pre_pack: enumerating states to pack")
            cursor = store_connection.cursor
            stmt = "%(TRUNCATE)s pack_state"
            self.runner.run_script_stmt(cursor, stmt)
            to_remove = 0


            if self.options.pack_gc:
                # Mark all objects we said not to keep as something
                # we should discard.
                stmt = """
                INSERT INTO pack_state (tid, zoid)
                SELECT tid, zoid
                FROM object_state
                INNER JOIN pack_object USING (zoid)
                WHERE keep = %(FALSE)s
                    AND tid > 0
                    AND tid <= %(pack_tid)s
                ORDER BY zoid
                """
                self.runner.run_script_stmt(
                    cursor, stmt, {'pack_tid': pack_tid})
                to_remove += cursor.rowcount
            else:
                # Support for IExternalGC. Also remove deleted objects.
                stmt = """
                INSERT INTO pack_state (tid, zoid)
                SELECT t.tid, t.zoid
                FROM (
                    SELECT zoid, tid
                    FROM object_state
                    WHERE state IS NULL
                    AND tid = (
                        SELECT MAX(i.tid)
                        FROM object_state i
                        WHERE i.zoid = object_state.zoid
                    )
                ) t
                """
                self.runner.run_script_stmt(cursor, stmt)
                to_remove += cursor.rowcount

            # Pack object states with the keep flag set to true,
            # excluding their current TID.
            stmt = """
            INSERT INTO pack_state (tid, zoid)
            SELECT tid, zoid
            FROM object_state
            INNER JOIN pack_object USING (zoid)
            WHERE keep = %(TRUE)s
                AND tid > 0
                AND tid != keep_tid
                AND tid <= %(pack_tid)s
            ORDER BY zoid
            """
            self.runner.run_script_stmt(
                cursor, stmt, {'pack_tid': pack_tid})
            to_remove += cursor.rowcount

            # Make a simple summary of the transactions to examine.
            logger.info("pre_pack: enumerating transactions to pack")
            stmt = "%(TRUNCATE)s pack_state_tid"
            self.runner.run_script_stmt(cursor, stmt)
            stmt = """
            INSERT INTO pack_state_tid (tid)
            SELECT DISTINCT tid
            FROM pack_state
            """
            cursor.execute(stmt)

            logger.info("pre_pack: will remove %d object state(s)",
                        to_remove)

            logger.info("pre_pack: finished successfully")
            store_connection.commit()
        except:
            store_connection.rollback_quietly()
            raise
        finally:
            store_connection.drop()
            load_connection.drop()

    def __initial_populate_pack_object(self, load_connection, store_connection,
                                       pack_tid, keep):
        """
        Put all objects into ``pack_object`` that have revisions equal
        to or below *pack_tid*, setting their initial ``keep`` status
        to *keep*.

        Commits the transaction to release locks.
        """
        # Access the tables that are used by online transactions
        # in a short transaction and immediately commit to release any
        # locks.

        # TRUNCATE may or may not cause implicit commits. (MySQL: Yes,
        # PostgreSQL: No)
        self.runner.run_script(store_connection.cursor, "%(TRUNCATE)s pack_object;")

        affected_objects = """
        SELECT zoid, tid
        FROM object_state
        WHERE tid > 0 AND tid <= %(pack_tid)s
        ORDER BY zoid
        """

        # Take the locks we need up front, in order, because
        # locking in a subquery doing an INSERT isn't guaranteed to use that
        # order (deadlocks seen with commits on MySQL 5.7 without this,
        # when using REPEATABLE READ.)
        #
        # We must do this on its own, because some drivers (notably
        # mysql-connector-python) get very upset
        # ("mysql.connector.errors.InternalError: Unread result
        # found") if you issue a SELECT that you don't then consume.
        #
        # Since we switched MySQL back to READ COMMITTED (what PostgreSQL uses)
        # I haven't been able to produce the error anymore. So don't explicitly lock.

        stmt = """
        INSERT INTO pack_object (zoid, keep, keep_tid)
        SELECT zoid, """ + ('%(TRUE)s' if keep else '%(FALSE)s') + """, MAX(tid)
        FROM ( """ + affected_objects + """ ) t
        GROUP BY zoid;

        -- Keep the root object.
        UPDATE pack_object
        SET keep = %(TRUE)s
        WHERE zoid = 0;
        """
        self.runner.run_script(store_connection.cursor, stmt, {'pack_tid': pack_tid})
        store_connection.commit()

    def _pre_pack_without_gc(self, conn, cursor, pack_tid):
        """
        Determine what to pack, without garbage collection.

        With garbage collection disabled, there is no need to follow
        object references.
        """
        # Fill the pack_object table with OIDs, but configure them
        # all to be kept by setting keep to true.
        logger.debug("pre_pack: populating pack_object")
        self.__initial_populate_pack_object(conn, cursor, pack_tid, keep=True)

    def _pre_pack_with_gc(self, load_connection, store_connection,
                          pack_tid, get_references):
        """
        Determine what to pack, with garbage collection.
        """
        stmt = self._script_create_temp_pack_visit
        if stmt:
            self.runner.run_script(store_connection.cursor, stmt)

        self.fill_object_refs(load_connection, store_connection, get_references)

        logger.info("pre_pack: filling the pack_object table")
        # Fill the pack_object table with OIDs that either will be
        # removed (if nothing references the OID) or whose history will
        # be cut.
        self.__initial_populate_pack_object(load_connection, store_connection,
                                            pack_tid, keep=False)

        stmt = """
        -- Keep objects that have been revised since pack_tid.
        -- Use temp_pack_visit for temporary state; otherwise MySQL 5 chokes.
        INSERT INTO temp_pack_visit (zoid, keep_tid)
        SELECT zoid, 0
        FROM current_object
        WHERE tid > %(pack_tid)s
        ORDER BY zoid;

        UPDATE pack_object
        SET keep = %(TRUE)s
        WHERE zoid IN (
            SELECT zoid
            FROM temp_pack_visit
        );

        %(TRUNCATE)s temp_pack_visit;

        -- Keep objects that are still referenced by object states in
        -- transactions that will not be packed.
        -- Use temp_pack_visit for temporary state; otherwise MySQL 5 chokes.
        INSERT INTO temp_pack_visit (zoid, keep_tid)
        SELECT DISTINCT to_zoid, 0
        FROM object_ref
        WHERE tid > %(pack_tid)s;

        UPDATE pack_object
        SET keep = %(TRUE)s
        WHERE zoid IN (
            SELECT zoid
            FROM temp_pack_visit
        );

        %(TRUNCATE)s temp_pack_visit;
        """
        self.runner.run_script(store_connection.cursor, stmt, {'pack_tid': pack_tid})

        # Traverse the graph, setting the 'keep' flags in pack_object
        self._traverse_graph(load_connection, store_connection)
        store_connection.commit()

    def _find_pack_tid(self):
        """If pack was not completed, find our pack tid again"""
        conn, cursor = self.connmanager.open_for_pre_pack()
        try:
            stmt = self._choose_pack_tid_query
            stmt.execute(cursor)
            res = [tid for (tid,) in cursor]
        finally:
            self.connmanager.close(conn, cursor)
        return res[0] if res else 0


    @metricmethod
    def pack(self, pack_tid, packed_func=None):
        """Pack.  Requires the information provided by pre_pack."""
        # pylint:disable=too-many-locals
        # Read committed mode is sufficient.

        conn, cursor = self.connmanager.open_for_store()
        try: # pylint:disable=too-many-nested-blocks
            try:
                # If we have a transaction entry in ``pack_state_tid`` (that is,
                # we found a transaction with an object in the range of transactions
                # we can pack away) that matches an actual transaction entry (XXX:
                # How could we be in the state where the transaction row is gone but we still
                # have object_state with that transaction id?), then we need to pack that
                # transaction. The presence of an entry in ``pack_state_tid`` means that all
                # object states from that transaction should be removed.
                stmt = """
                SELECT tx.tid,
                       CASE WHEN packed = %(TRUE)s THEN 1 ELSE 0 END,
                       CASE WHEN pack_state_tid.tid IS NOT NULL THEN 1 ELSE 0 END
                FROM "transaction" tx
                LEFT OUTER JOIN pack_state_tid ON (tx.tid = pack_state_tid.tid)
                WHERE tx.tid > 0
                    AND tx.tid <= %(pack_tid)s
                    AND (packed = %(FALSE)s OR pack_state_tid.tid IS NOT NULL)
                ORDER BY tx.tid
                """
                self.runner.run_script_stmt(
                    cursor, stmt, {'pack_tid': pack_tid})
                tid_rows = list(cursor) # oldest first, sorted in SQL

                total = len(tid_rows)
                logger.info("pack: will pack %d transaction(s)", total)

                stmt = self._script_create_temp_pack_visit
                if stmt:
                    self.runner.run_script(cursor, stmt)

                # Lock and delete rows in the same order that
                # new commits would in order to prevent deadlocks.
                # Pack in small batches of transactions only after we are able
                # to obtain a commit lock in order to minimize the
                # interruption of concurrent write operations.
                start = time.time()
                packed_list = []
                counter, lastreport, statecounter = 0, 0, 0
                # We'll report on progress in at most .1% step increments
                reportstep = max(total / 1000, 1)

                for tid, packed, has_removable in tid_rows:
                    self._pack_transaction(
                        cursor, pack_tid, tid, packed, has_removable,
                        packed_list)
                    counter += 1
                    if time.time() >= start + self.options.pack_batch_timeout:
                        self.connmanager.commit(conn, cursor)
                        if packed_func is not None:
                            for poid, ptid in packed_list:
                                packed_func(poid, ptid)
                        statecounter += len(packed_list)
                        if counter >= lastreport + reportstep:
                            logger.info("pack: packed %d (%.1f%%) transaction(s), "
                                        "affecting %d states",
                                        counter, counter / float(total) * 100,
                                        statecounter)
                            lastreport = counter / reportstep * reportstep
                        del packed_list[:]
                        start = time.time()
                if packed_func is not None:
                    for oid, tid in packed_list:
                        packed_func(oid, tid)
                packed_list = None

                self._pack_cleanup(conn, cursor)

            except:
                logger.exception("pack: failed")
                self.connmanager.rollback_quietly(conn, cursor)
                raise

            else:
                logger.info("pack: finished successfully")
                self.connmanager.commit(conn, cursor)

        finally:
            self.connmanager.close(conn, cursor)


    def _pack_transaction(self, cursor, pack_tid, tid, packed,
                          has_removable, packed_list):
        """
        Pack one transaction. Requires populated pack tables.

        If *has_removable* is true, then we have object states and current
        object pointers to remove.
        """
        logger.debug("pack: transaction %d: packing", tid)
        removed_objects = 0
        removed_states = 0

        if has_removable:
            stmt = self._script_pack_current_object
            self.runner.run_script_stmt(cursor, stmt, {'tid': tid})
            removed_objects = cursor.rowcount

            stmt = self._script_pack_object_state
            self.runner.run_script_stmt(cursor, stmt, {'tid': tid})
            removed_states = cursor.rowcount

            # Terminate prev_tid chains
            stmt = """
            UPDATE object_state SET prev_tid = 0
            WHERE prev_tid = %(tid)s
                AND tid <= %(pack_tid)s
            """
            self.runner.run_script_stmt(cursor, stmt,
                                        {'pack_tid': pack_tid, 'tid': tid})

            stmt = """
            SELECT pack_state.zoid
            FROM pack_state
            WHERE pack_state.tid = %(tid)s
            """
            self.runner.run_script_stmt(cursor, stmt, {'tid': tid})
            for (oid,) in cursor:
                packed_list.append((oid, tid))

        # Find out whether the transaction is empty
        stmt = self._transaction_has_data_query
        stmt.execute(cursor, {'tid': tid})
        empty = not list(cursor)

        # mark the transaction packed and possibly empty
        if empty:
            clause = 'is_empty = %(TRUE)s'
            state = 'empty'
        else:
            clause = 'is_empty = %(FALSE)s'
            state = 'not empty'
        stmt = 'UPDATE "transaction" SET packed = %(TRUE)s, ' + clause
        stmt += " WHERE tid = %(tid)s"
        self.runner.run_script_stmt(cursor, stmt, {'tid': tid})

        logger.debug(
            "pack: transaction %d (%s): removed %d object(s) and %d state(s)",
            tid, state, removed_objects, removed_states)


    def _pack_cleanup(self, conn, cursor):
        """Remove unneeded table rows after packing"""
        # commit the work done so far, releasing row-level locks.
        self.connmanager.commit(conn, cursor)
        logger.info("pack: cleaning up")

        # This section does not need to hold the commit lock, as it only
        # touches pack-specific tables. We already hold a pack lock for that.
        logger.debug("pack: removing unused object references")
        stmt = self._script_pack_object_ref
        self.runner.run_script(cursor, stmt)

        # We need a commit lock when touching the transaction table though.
        # We'll do it in batches of 1000 rows.
        logger.debug("pack: removing empty packed transactions")
        while True:
            stmt = self._delete_empty_transactions_batch_query
            stmt.execute(cursor)
            deleted = cursor.rowcount
            self.connmanager.commit(conn, cursor)
            self.locker.release_commit_lock(cursor)
            if deleted < 1000:
                # Last set of deletions complete
                break

        # perform cleanup that does not require the commit lock
        logger.debug("pack: clearing temporary pack state")
        for _table in ('pack_object', 'pack_state', 'pack_state_tid'):
            stmt = '%(TRUNCATE)s ' + _table
            self.runner.run_script_stmt(cursor, stmt)


@implementer(IPackUndo)
class HistoryFreePackUndo(PackUndo):
    """
    History-free pack/undo.
    """

    keep_history = False

    # How often, in seconds, to commit work in progress.
    # This is a variable here for testing.
    fill_object_refs_commit_frequency = 60

    # How many object states to find references in at any one time.
    # This is a control on the amount of memory used by the Python
    # process during packing, especially if the database driver
    # doesn't use server-side cursors.
    fill_object_refs_batch_size = 100

    _choose_pack_transaction_query = Schema.object_state.select(
        it.c.tid
    ).where(
        it.c.tid > 0
    ).and_(
        it.c.tid <= it.bindparam('tid')
    ).order_by(
        it.c.tid, 'DESC'
    ).limit(1)

    # history-free packing doesn't use temp_pack_visit, unlike
    # history-preserving.
    # _script_create_temp_pack_visit = """
    #     CREATE TEMPORARY TABLE temp_pack_visit (
    #         zoid BIGINT NOT NULL PRIMARY KEY,
    #         keep_tid BIGINT NOT NULL
    #     );
    #     CREATE INDEX temp_pack_keep_tid ON temp_pack_visit (keep_tid)
    #     """

    _script_delete_object = """
    DELETE FROM object_state
    WHERE zoid = %(oid)s
    and tid = %(tid)s
    """

    def on_fill_object_ref_batch(self, oid_batch, refs_found):
        """Hook for testing."""

    def verify_undoable(self, cursor, undo_tid):
        """Raise UndoError if it is not safe to undo the specified txn."""
        raise UndoError("Undo is not supported by this storage")

    def undo(self, cursor, undo_tid, self_tid):
        """Undo a transaction.

        Parameters: "undo_tid", the integer tid of the transaction to undo,
        and "self_tid", the integer tid of the current transaction.

        Returns the list of OIDs undone.
        """
        raise UndoError("Undo is not supported by this storage")

    def fill_object_refs(self, load_connection, store_connection, get_references):
        """
        Update the object_refs table by analyzing new object states.

        See :meth:`pre_pack` for a description of the parameters.

        Because *load_connection* is read-only and repeatable read,
        we don't need to do any object-level locking.
        """
        # Begin by ensuring we have a snapshot reflecting anything
        # committed up to this point, including the contents of
        # ``pack_object``, which determines the visible objects
        # we will examine.
        load_connection.restart()
        mem_begin = get_memory_usage()
        logger.debug("pre_pack: Collecting objects to examine.")
        # Recall pre_pack can be run many times.
        # Ordering should be immaterial as we are in a read-only snapshot view
        # of the database; we shouldn't run into locking issues with other
        # transactions.
        with self._make_ss_load_cursor(load_connection) as ss_load_cursor:
            stmt = """
            SELECT zoid
            FROM pack_object
            INNER JOIN object_state USING (zoid)
            LEFT OUTER JOIN object_refs_added
                USING (zoid)
            WHERE object_refs_added.tid IS NULL
              OR object_refs_added.tid != object_state.tid
            """
            ss_load_cursor.execute(stmt)
            logger.debug(
                "pre_pack: Selected objects to examine (memory delta: %s)",
                byte_display(get_memory_usage() - mem_begin)
            )

            oids = OidList((row[0] for row in ss_load_cursor))

        log_at = time.time() + self.fill_object_refs_commit_frequency
        self.on_filling_object_refs_added(oids=oids)

        oid_count = len(oids)
        oids_done = 0
        num_refs_found = 0
        # Against 30 MM rows with MySQL on Python 2.7 with the
        # server-side cursor on mysqlclient, the SELECT takes
        # negligible time and memory; transforming into the
        # array.array and pulling rows takes 5 minutes and a total of
        # 231MB with a batch size of 1024; that's about the resident
        # size of the process, too. Using a fetch size of 10000 didn't reduce
        # the time substantially though curiously it did report just 165.9MB
        # memory delta.
        #
        # Previously, using a list and a buffered cursor for the same setup,
        # the SELECT takes 3.5 minutes and a memory delta of 2.5GB; transforming into
        # the list takes a few more seconds and a final delta of 3GB. When the cursor
        # is closed, the resident size of the process shrinks to around 1.2GB.
        logger.info(
            "pre_pack: analyzing references from %d object(s) (memory delta: %s)",
            oid_count, byte_display(get_memory_usage() - mem_begin))
        while oids_done < oid_count:
            # Previously, we iterated like this:
            #
            # while oids:
            #    batch = oids[:batch_size]
            #    oids =  oids[batch_size:]
            #
            # However, that turns into O(n^2) operations with a large
            # overhead, especially on CPython. Each slice operation
            # allocates a new list, and copies into it (including
            # INCREF). Even with the O(n^2) algorithm,
            # array.array('Q') benchmarks ~5x faster for these two
            # operations, probably because it's just memory movements,
            # not loops that have to INCREF/DECREF.
            #
            # A simple profile while this is running with 30 MM rows
            # shows at least 37% of the time spent in the C
            # ``list_slice`` function and 31% in ``list_dealloc``,
            # averaging about 15,000 objects per minute.
            #
            # Switching just to array.array and leaving the slicing, I
            # was getting 44,000 objects per minute, but 99% time
            # spent in memmove().
            #
            # Using manual indexing of arrays, CPU usage of less than
            # 35%; for the first time, 35% of profile time is spent in
            # talking to MySQL (over gigabit switch); clearly parallel
            # pre-fetching would be useful.

            batch = oids[oids_done:oids_done + self.fill_object_refs_batch_size]
            oids_done += len(batch)

            refs_found = self._add_refs_for_oids(load_connection, store_connection,
                                                 batch, get_references)
            num_refs_found += len(refs_found)
            self.on_fill_object_ref_batch(oid_batch=batch, refs_found=refs_found)

            now = time.time()
            if now >= log_at:
                # Save the work done so far.
                store_connection.commit()
                log_at = now + self.fill_object_refs_commit_frequency
                logger.info(
                    "pre_pack: objects analyzed: %d/%d (%d total references)",
                    oids_done, oid_count, num_refs_found)
        # Those 30MM objects wound up with about 48,976,835 references.
        store_connection.commit()
        logger.info(
            "pre_pack: objects analyzed: %d/%d", oids_done, oid_count)

    def _add_refs_for_oids(self, load_connection, store_connection,
                           oids, get_references):
        """Fill object_refs with the states for some objects.

        Returns the number of references added.
        """
        # TODO: Use the row batcher's SELECT FROM
        load_cursor = load_connection.cursor
        oid_list = ','.join(str(oid) for oid in oids)
        stmt = """
            SELECT zoid, tid, state
            FROM object_state
            WHERE zoid IN (%s)
            ORDER BY zoid
            """ % oid_list
        self.runner.run_script_stmt(load_cursor, stmt)

        add_objects = []
        add_refs = []

        for from_oid, tid, state in load_cursor:
            state = self.driver.binary_column_as_state_type(state)
            add_objects.append((from_oid, tid))
            if state:
                try:
                    to_oids = get_references(state)
                except:
                    logger.exception(
                        "pre_pack: can't unpickle "
                        "object %d in transaction %d; state length = %d",
                        from_oid, tid, len(state)
                    )
                    raise
                for to_oid in to_oids:
                    add_refs.append((from_oid, tid, to_oid))
        if not add_objects:
            assert not add_refs
            return add_refs

        # TODO: RowBatcher for all of these
        store_cursor = store_connection.cursor
        stmt = "DELETE FROM object_refs_added WHERE zoid IN (%s)" % oid_list
        self.runner.run_script_stmt(store_cursor, stmt)
        stmt = "DELETE FROM object_ref WHERE zoid IN (%s)" % oid_list
        self.runner.run_script_stmt(store_cursor, stmt)

        stmt = """
        INSERT INTO object_ref (zoid, tid, to_zoid) VALUES (%s, %s, %s)
        """
        self.runner.run_many(store_cursor, stmt, add_refs)

        stmt = """
        INSERT INTO object_refs_added (zoid, tid) VALUES (%s, %s)
        """
        self.runner.run_many(store_cursor, stmt, add_objects)

        return add_refs

    @metricmethod
    def pre_pack(self, pack_tid, get_references):
        """
        Decide what the garbage collector should delete.

        Objects created or modified after pack_tid will not be garbage
        collected.

        get_references is a function that accepts a pickled state and
        returns a set of OIDs that state refers to.

        The self.options.pack_gc flag indicates whether to run garbage
        collection. If pack_gc is false, this method does nothing.
        """
        if not self.options.pack_gc:
            logger.warning("pre_pack: garbage collection is disabled on a "
                           "history-free storage, so doing nothing")
            return

        load_connection = LoadConnection(self.connmanager)
        store_connection = PrePackConnection(self.connmanager)
        try:
            try:
                self._pre_pack_main(load_connection, store_connection,
                                    pack_tid, get_references)
            except:
                logger.exception("pre_pack: failed")
                store_connection.rollback_quietly()
                raise
            else:
                store_connection.commit()
                logger.info("pre_pack: finished successfully")
        finally:
            load_connection.drop()
            store_connection.drop()


    def _pre_pack_main(self, load_connection, store_connection,
                       pack_tid, get_references):
        """
        Determine what to garbage collect.

        *load_connection* is a
        :class:`relstorage.adapters.connections.LoadConnection`; this
        connection is in "snapshot" mode and is used to read a
        consistent view of the database. Although this connection is
        never committed or rolled back while this method is running
        (which may take a long time), because load connections are
        declared to be read-only the database engines can make certain
        optimizations that reduce the overhead of them (e.g.,
        https://dev.mysql.com/doc/refman/5.7/en/innodb-performance-ro-txn.html),
        making long-running transactions less problematic. For
        example, while packing a 60 million row single MySQL storage
        with ``zc.zodbdgc``, a load transaction was open and actively
        reading for over 8 hours while the database continued to be
        heavily written to without causing any problems.

        *store_connection* is a standard read-committed store connection;
        it will be periodically committed.
        """
        # First, fill the ``pack_object`` table with all known OIDs
        # as they currently exist in the database, regardless of
        # what the load_connection snapshot can see (which is no later
        # and possibly earlier, than what the store connection can see).
        #
        # Mark things that need to be kept:
        # - the root object;
        # - anything that has changed since ``pack_tid``;
        # Note that we do NOT add items that have been newly added since
        # ``pack_tid``; no need to traverse into them, they couldn't possibly
        # have a reference to an older object that's not also referenced
        # by an object in the snapshot (without the app doing something seriously
        # wrong): plus, we didn't find references from that item anyway.
        #
        # TODO: Copying 30MM objects takes almost 10 minutes (600s)
        # against mysql 8 running on an SSD, and heaven forgive you if
        # you kill the transaction and roll back --- the undo info is
        # insane. What if we CREATE AS SELECT a table? Doing 'CREATE
        # TEMPORARY TABLE AS' takes 173s; doing 'CREATE TABLE AS'
        # takes 277s.
        #
        # On PostgreSQL we could use unlogged tables.
        logger.info("pre_pack: filling the pack_object table")
        stmt = """
        %(TRUNCATE)s pack_object;

        INSERT INTO pack_object (zoid, keep, keep_tid)
        SELECT zoid, CASE WHEN tid > %(pack_tid)s THEN %(TRUE)s ELSE %(FALSE)s END, tid
        FROM object_state;

        -- Also keep the root
        UPDATE pack_object
        SET keep = %(TRUE)s
        WHERE zoid = 0;
        """
        self.runner.run_script(store_connection.cursor, stmt, {'pack_tid': pack_tid})
        store_connection.commit()
        logger.info("pre_pack: Filled the pack_object table")

        # Chase down all the references using a consistent snapshot, including
        # only the objects that were visible in ``pack_object``.
        self.fill_object_refs(load_connection, store_connection, get_references)

        # Traverse the graph, setting the 'keep' flags in ``pack_object``
        self._traverse_graph(load_connection, store_connection)


    def _find_pack_tid(self):
        """If pack was not completed, find our pack tid again"""

        # pack (below) ignores its pack_tid argument, so we can safely
        # return None here
        return None


    @metricmethod
    def pack(self, pack_tid, packed_func=None):
        """Run garbage collection.

        Requires the information provided by pre_pack.
        """
        # pylint:disable=too-many-locals
        # Read committed mode is sufficient.
        conn, cursor = self.connmanager.open_for_store()
        try: # pylint:disable=too-many-nested-blocks
            try:
                stmt = """
                SELECT zoid, keep_tid
                FROM pack_object
                WHERE keep = %(FALSE)s
                ORDER BY zoid
                """
                self.runner.run_script_stmt(cursor, stmt)
                to_remove = list(cursor)

                total = len(to_remove)
                logger.info("pack: will remove %d object(s)", total)

                # Hold the commit lock while packing to prevent deadlocks.
                # Pack in small batches of transactions only after we are able
                # to obtain a commit lock in order to minimize the
                # interruption of concurrent write operations.
                start = time.time()
                packed_list = []
                # We'll report on progress in at most .1% step increments
                lastreport, reportstep = 0, max(total / 1000, 1)

                while to_remove:
                    # TODO: Use the row batcher for this,
                    # or simply do a join against the table.
                    items = to_remove[:100]
                    del to_remove[:100]
                    # XXX: History free. We shouldn't need to include the TID,
                    # but that's probably there to protect against concurrent modification
                    # of the object. With our row-level locking in 3.0 this may not strictly be
                    # necessary?
                    stmt = """
                    DELETE FROM object_state
                    WHERE zoid = %s AND tid = %s
                    """
                    self.runner.run_many(cursor, stmt, items)
                    packed_list.extend(items)

                    if time.time() >= start + self.options.pack_batch_timeout:
                        self.connmanager.commit(conn, cursor)
                        if packed_func is not None:
                            for oid, tid in packed_list:
                                packed_func(oid, tid)
                        del packed_list[:]
                        counter = total - len(to_remove)
                        if counter >= lastreport + reportstep:
                            logger.info("pack: removed %d (%.1f%%) state(s)",
                                        counter, counter / float(total) * 100)
                            lastreport = counter / reportstep * reportstep
                        start = time.time()

                if packed_func is not None:
                    for oid, tid in packed_list:
                        packed_func(oid, tid)
                packed_list = None

                self._pack_cleanup(conn, cursor)

            except:
                logger.exception("pack: failed")
                self.connmanager.rollback_quietly(conn, cursor)
                raise

            else:
                logger.info("pack: finished successfully")
                self.connmanager.commit(conn, cursor)
        finally:
            self.connmanager.close(conn, cursor)


    def _pack_cleanup(self, conn, cursor):
        # commit the work done so far
        self.connmanager.commit(conn, cursor)
        self.locker.release_commit_lock(cursor)
        logger.info("pack: cleaning up")

        # This section does not need to hold the commit lock, as it only
        # touches pack-specific tables. We already hold a pack lock for that.
        # XXX: Shouldn't we keep this? Unless there's a huge amount of churn,
        # older state info might be valid from previous packs. I guess we
        # want to avoid backing these things up.
        stmt = """
        DELETE FROM object_refs_added
        WHERE zoid IN (
            SELECT zoid
            FROM pack_object
            WHERE keep = %(FALSE)s
        );

        DELETE FROM object_ref
        WHERE zoid IN (
            SELECT zoid
            FROM pack_object
            WHERE keep = %(FALSE)s
        );

        %(TRUNCATE)s pack_object
        """
        self.runner.run_script(cursor, stmt)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import, print_function


from zope.interface import implementer

from .._compat import metricmethod
from .interfaces import IConnectionManager
from .interfaces import ReplicaClosedException
from .replica import ReplicaSelector

logger = __import__('logging').getLogger(__name__)


def connection_callback(isolation_level=None,
                        read_only=None,
                        deferrable=None,
                        application_name=None,
                        inherit=None):
    """
    Decorator for functions used in :meth:`IConnectionManager.open_and_call`.

    When the function is called, it will be given a connection that has
    the given traits.

    Use ``inherit=BaseClass.func`` when you're overriding a callback
    function ``func``.
    """
    if inherit is not None:
        isolation_level = isolation_level or inherit.transaction_isolation_level
        application_name = application_name or inherit.transaction_application_name
        read_only = read_only or inherit.transaction_read_only
        deferrable = deferrable or inherit.transaction_deferrable
    def f(func):
        func.transaction_isolation_level = isolation_level
        func.transaction_read_only = read_only
        func.transaction_deferrable = deferrable
        func.transaction_application_name = application_name
        return func
    return f


def _connection_callback_open_args(connmanager, callback):
    # Read the settings from `connection_callback` or defaults.
    isolation = getattr(callback, 'transaction_isolation_level', None)
    read_only = getattr(callback, 'transaction_read_only', None)
    deferrable = getattr(callback, 'transaction_deferrable', None)
    application_name = getattr(callback, 'transaction_application_name', None)
    return dict(
        isolation=isolation or connmanager.isolation_store,
        read_only=read_only or False,
        deferrable=deferrable or False,
        application_name=application_name or ('RS: ' + callback.__name__)
    )


@implementer(IConnectionManager)
class AbstractConnectionManager(object):
    """
    Abstract base class for connection management.

    Responsible for opening and closing database connections.
    """

    # a series of callables (cursor, restart=bool)
    # for when a store connection is opened.
    _on_store_opened = ()

    # a series of callables (cursor,) that will be called
    # when a load connection is opened
    _on_load_opened = ()

    # The list of exceptions to ignore on a rollback *or* close. We
    # take this as the union of the driver's close exceptions and disconnected
    # exceptions (drivers aren't required to organize them to overlap, but
    # in practice they should.)
    _ignored_exceptions = ()

    # Subclasses should set these to get semantics as close
    # as possible to these standard levels.
    isolation_serializable = None
    isolation_read_committed = None
    # If these are not set by a subclass, they will be copied
    # from isolation_serializable and read_committed, respectively.
    isolation_load = None
    isolation_store = None

    replica_selector = None
    ro_replica_selector = None

    def __init__(self, options, driver):
        """
        :param options: A :class:`relstorage.options.Options`.
        :param driver: A :class:`relstorage.adapters.interfaces.IDBDriver`,
            which we use for its exceptions.
        """
        self.keep_history = options.keep_history
        self.driver = driver
        self.options = options

        self._ignored_exceptions = tuple(set(
            driver.close_exceptions
            + driver.disconnected_exceptions
            + (ReplicaClosedException,)
        ))

        if options.replica_conf:
            self.replica_selector = ReplicaSelector(
                options.replica_conf, options.replica_timeout)

        if options.ro_replica_conf:
            self.ro_replica_selector = ReplicaSelector(
                options.ro_replica_conf, options.replica_timeout)
        else:
            self.ro_replica_selector = self.replica_selector

        if not self.isolation_load:
            self.isolation_load = self.isolation_serializable
        if not self.isolation_store:
            self.isolation_store = self.isolation_read_committed

        self._may_need_rollback = driver.connection_may_need_rollback
        self._may_need_commit = driver.connection_may_need_commit
        self._synchronize_cursor_for_rollback = driver.synchronize_cursor_for_rollback
        self._do_commit = driver.commit
        self._do_rollback = driver.rollback

    def add_on_store_opened(self, f):
        """
        Add a callable(cursor, restart=bool) for when a store connection
        is opened.

        Hooks are called in the order added.

        .. versionadded:: 2.1a1
        """
        self._on_store_opened += (f,)

    set_on_store_opened = add_on_store_opened # BWC

    def add_on_load_opened(self, f):
        """
        Add a callable (cursor, restart=bool) for when a load connection is opened.

        Hooks are called in the order added.

        .. versionadded:: 2.1a1
        """
        self._on_load_opened += (f,)

    def open(self,
             isolation=None,
             read_only=False,
             deferrable=False,
             replica_selector=None,
             application_name=None,
             **kwargs):
        """Open a database connection and return (conn, cursor)."""
        raise NotImplementedError()

    @metricmethod
    def close(self, conn=None, cursor=None):
        """
        Close a connection and cursor, ignoring certain errors.

        Return a True value if the connection was closed cleanly. Return
        a False value if the processes ignored an error.
        """
        clean = True
        for obj in (cursor, conn): # cursor first; some drivers want that done
            if obj is not None:
                try:
                    obj.close()
                except self._ignored_exceptions: # pylint:disable=catching-non-exception
                    clean = False
        return clean

    def __rollback_connection(self, conn, ignored_exceptions, restarting):
        """Return True if we successfully rolled back."""
        clean = True
        if conn is not None:
            if self._may_need_rollback(conn):
                try:
                    self._do_rollback(conn)
                except ignored_exceptions:
                    clean = False
            elif restarting:
                self._begin_for_restart(conn)
        return clean

    def _begin_for_restart(self, conn):
        pass

    def _may_need_rollback(self, conn): # pylint:disable=unused-argument,method-hidden
        """
        Answer if this connection might need to be rolled back.

        If a subclass can definitively say that it does *not* need to
        be rolled back, because it is not in a transaction,
        it can override to return false.
        """
        return True

    def _may_need_commit(self, conn): # pylint:disable=unused-argument,method-hidden
        """
        Answer if this connection might need to be committed.

        If a subclass can definitively say that it does *not* need to
        be committed, because it is not in a transaction,
        it can override to return false.
        """
        return True

    def __rollback(self, conn, cursor, quietly, restarting):
        # If an error occurs, close the connection and cursor.
        #
        # Some drivers require the cursor to be closed before closing
        # the connection.
        #
        # Some drivers also don't allow you to close the cursor
        # without fetching all rows.
        self._synchronize_cursor_for_rollback(cursor)
        try:
            clean = self.__rollback_connection(
                conn,
                # Let it raise if we're not meant to be quiet.
                self._ignored_exceptions if quietly else (),
                restarting
            )
        except:
            clean = False
            raise
        finally:
            if not clean:
                self.close(conn, cursor)
        return clean

    def rollback_and_close(self, conn, cursor):
        clean = self.__rollback(conn, cursor, True, False)
        if clean:
            # if an error already occurred, we closed things.
            clean = self.close(conn, cursor)

        return clean

    def rollback(self, conn, cursor):
        return self.__rollback(conn, cursor, False, None)

    def rollback_for_restart(self, conn, cursor):
        return self.__rollback(conn, cursor, False, True)

    def rollback_quietly(self, conn, cursor):
        return self.__rollback(conn, cursor, True, None)

    rollback_store_quietly = rollback_quietly

    def commit(self, conn, cursor=None, force=False):
        if self._may_need_commit(conn) or force:
            self._do_commit(conn, cursor)

    def begin(self, conn, cursor):
        pass

    def cursor_for_connection(self, conn):
        return self.driver.cursor(conn)

    def open_and_call(self, callback):
        """
        Call ``callback(connection, cursor)`` with a newly open connection and cursor.

        If the function returns, commits the transaction and returns the
        result returned by the function.
        If the function raises an exception, aborts the transaction
        then propagates the exception.
        """
        conn, cursor = self._do_open_for_call(callback)
        try:
            try:
                res = callback(conn, cursor)
            except:
                self.rollback_and_close(conn, cursor)
                conn, cursor = None, None
                raise
            else:
                self.close(None, cursor)
                cursor = None
                self.commit(conn)
                return res
        finally:
            self.close(conn, cursor)

    def _call_hooks(self, hooks, conn, cursor,
                    *args, **kwargs):
        try:
            for hook in hooks:
                hook(*args, **kwargs)
        except:
            self.close(conn, cursor)
            raise

    def _do_open_for_load(self):
        raise NotImplementedError()

    def open_for_load(self):
        conn, cursor = self._do_open_for_load()
        self._call_hooks(self._on_load_opened, conn, cursor,
                         cursor, restart=False)
        return conn, cursor

    def restart_load(self, conn, cursor, needs_rollback=True):
        """Reinitialize a connection for loading objects."""
        self.check_replica(conn, cursor,
                           replica_selector=self.ro_replica_selector)
        if needs_rollback:
            self.rollback(conn, cursor)
        self._call_hooks(self._on_load_opened, conn, cursor,
                         cursor, restart=True)

    def check_replica(self, conn, cursor, replica_selector=None):
        """Raise an exception if the connection belongs to an old replica"""
        if replica_selector is None:
            replica_selector = self.replica_selector

        if replica_selector is not None:
            current = replica_selector.current()
            if conn.replica != current:
                # Prompt the change to a new replica by raising an exception.
                self.close(conn, cursor)
                raise ReplicaClosedException(
                    "Switched replica from %s to %s" % (conn.replica, current))

    def open_for_pre_pack(self):
        """Open a connection to be used for the pre-pack phase.
        Returns (conn, cursor).
        """
        return self.open_for_store(application_name='RS: Prepack')

    def open_for_pack_lock(self):
        return self.open()

    def _do_open_for_store(self, **open_args):
        open_args['isolation'] = self.isolation_store
        open_args['read_only'] = False
        open_args['deferrable'] = False
        if 'application_name' not in open_args:
            open_args['application_name'] = 'RS: Store'
        return self.open(**open_args)

    def _do_open_for_call(self, callback):
        return self.open(**_connection_callback_open_args(self, callback))

    def _after_opened_for_store(self, conn, cursor, restart=False):
        """
        Called after a store is opened or restarted but
        before hooks are called.

        Subclasses may override.
        """
        # pylint:disable=unused-argument
        return

    def open_for_store(self, **open_args):
        """Open and initialize a connection for storing objects.

        Returns (conn, cursor).
        """
        conn, cursor = self._do_open_for_store(**open_args)
        try:
            self._after_opened_for_store(conn, cursor)
        except:
            self.close(conn, cursor)
            raise
        self._call_hooks(self._on_store_opened, conn, cursor,
                         cursor, restart=False)

        return conn, cursor

    def restart_store(self, conn, cursor, needs_rollback=True):
        """Reuse a store connection."""
        self.check_replica(conn, cursor)
        if needs_rollback:
            self.rollback_for_restart(conn, cursor)
        self._after_opened_for_store(conn, cursor)
        self._call_hooks(self._on_store_opened, conn, cursor,
                         cursor, restart=True)



##############################################################################
#
# Copyright (c) 2017 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"Internal helper utilities."
from __future__ import print_function
from __future__ import absolute_import

from collections import namedtuple
from functools import partial
from functools import update_wrapper
from functools import wraps

from .._compat import intern
from .._compat import MAX_TID

from .._util import Lazy

def query_property(base_name,
                   extension='',
                   formatted=False,
                   property_suffix='_queries',
                   lazy_suffix='_query'):
    """
    Defines a property that adapts to preserving or dropping history.

    To use, define a property ending in ``_queries``
    *property_suffix*) that is a two-tuple, where the preserving query
    comes first and the dropping query comes second. This indirection
    lets subclasses override these queries.

    Then define a property, passing the base name (without
    ``_queries``) to this function.

    The correct query will be lazily picked at runtime. The instance
    must have the ``keep_history`` attribute.

    If the chosen query is an exception instance or class, it will be
    raised instead of returned. This allows defining a query that is
    only supported in one of the two modes. Usually, this exception
    should be :class:`ZODB.POSException.Unsupported`.

    :keyword str extension: This string will be appended to whatever
        query is chosen before it is formatted and before it is returned.
    :keyword bool formatted: If True (*not* the default), then the
        chosen query will be formatted using the
        ``self.runner.script_vars``. This should produce a new
        query that uses the correct parameter binding for the database.
    """

    def prop(inst):
        queries = getattr(inst, base_name + property_suffix)
        query = queries[0] if inst.keep_history else queries[1]
        if isinstance(query, Exception) or (
                isinstance(query, type)
                and issubclass(query, Exception)):
            raise query

        if extension:
            query = query + extension
        if formatted:
            query = intern(query % inst.runner.script_vars)

        return query

    prop.__doc__ = "Query for " + base_name

    return Lazy(prop, base_name + lazy_suffix)

formatted_query_property = partial(query_property, formatted=True)


def noop_when_history_free(meth):
    """
    Decorator for *meth* that causes it to do nothing when
    ``self.keep_history`` is False.

    *meth* must have no return value (returns None) when it is
    history free. When history is preserved it can return anything.

    This requires a bit more memory to use the instance dict, but at
    runtime it has minimal time overhead (after the first call).
    """

    # Python 3.4 (via timeit)
    # calling a trivial method ('def t(self, arg): return arg') takes 118ns
    # calling a method that does 'if not self.keep_history: return; return arg'
    #   takes 142 ns
    # calling a functools.partial bound to self wrapped around t
    #   takes 298ns
    # calling a generic python function
    #     def wrap(self, *args, **kwargs):
    #       if not self.keep_history: return
    #       return self.t(*args, **kwargs)
    #   takes 429ns
    # So a partial function set into the __dict__ is the fastest way to
    # do this.

    meth_name = meth.__name__

    @wraps(meth)
    def no_op(*_args, **_kwargs):
        return

    @wraps(meth)
    def swizzler(self, *args, **kwargs):
        if not self.keep_history:
            setattr(self, meth_name, no_op)
        else:
            # NOTE: This creates a reference cycle
            bound = partial(meth, self)
            update_wrapper(bound, meth)
            if not hasattr(bound, '__wrapped__'):
                bound.__wrapped__ = meth
            setattr(self, meth_name, bound)

        return getattr(self, meth_name)(*args, **kwargs)

    if not hasattr(swizzler, '__wrapped__'):
        # Py2: this was added in 3.2
        swizzler.__wrapped__ = meth
        no_op.__wrapped__ = meth

    return swizzler


ResultDescription = namedtuple(
    'ResultDescription',
    # First two are mandatory, remaining five may be None
    # Example:
    # ('Name', 253, 17, 192, 192, 0, 0),
    ('name', 'type_code', 'display_size',
     'internal_size', 'precision', 'scale', 'null_ok'))


class DatabaseHelpersMixin(object):

    MAX_TID = MAX_TID

    def _metadata_to_native_str(self, value):
        # Some drivers, in some configurations, notably older versions
        # of MySQLdb (mysqlclient) on Python 3 in 'NAMES binary' mode,
        # can return column names and the like as bytes when we want native str.
        # pg8000 on Python2 does the reverse and returns unicode when we want
        # native str.
        if value is not None and not isinstance(value, str):
            # Checking for bytes tells us that it's a unicode object on Python 2;
            # we won't get here if it's bytes (because bytes is str)
            value = value.decode('ascii') if isinstance(value, bytes) else value.encode('ascii"')
        return value

    def _column_descriptions(self, cursor):
        __traceback_info__ = cursor.description
        return [ResultDescription(self._metadata_to_native_str(r[0]),
                                  # Not all drivers return lists or tuples
                                  # or things that can be sliced; psycopg2/cffi returns
                                  # an arbitrary sequence.
                                  # MySqlConnector-Python has been observed to provide
                                  # extra attributes.
                                  *list(r)[1:7])
                for r in cursor.description]

    def _rows_as_dicts(self, cursor):
        """
        An iterator of the rows as dictionaries, named by the
        lower-case column name.

        Some drivers offer the ability to do this directly when
        the statement is executed or the cursor is created;
        this is a lowest-common denominator way to do it utilizing
        DB-API 2.0 attributes.
        """
        column_descrs = self._column_descriptions(cursor)
        for row in cursor:
            result = {
                column_descr.name.lower(): column_value
                for column_descr, column_value in zip(column_descrs, row)
            }
            yield result

    def _rows_as_pretty_string(self, cursor):
        """
        Return the rows formatted in a way for easy human consumption.
        """
        from .._compat import NStringIO
        out = NStringIO()
        names = [d.name for d in self._column_descriptions(cursor)]
        kwargs = {'sep': '\t', 'file': out}
        print(*names, **kwargs)
        for row in cursor:
            print(*row, **kwargs)
        return out.getvalue()



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Batch table row insert/delete support.
"""
from collections import defaultdict
import itertools

from relstorage._compat import iteritems
from relstorage._compat import perf_counter
from relstorage._util import parse_byte_size

from .interfaces import AggregateOperationTimeoutError

# A cache
# {(placeholder, count): "string"}
_row_schemas = {}

class RowBatcher(object):
    """
    Generic row batcher.

    Inserting can use whatever paramater placeholder format
    and parameter format the cursor supports.

    Deleting needs to use ordered parameters. The placeholder
    can be set in the ``delete_placeholder`` attribute.
    """

    perf_counter = perf_counter

    # How many total rows can be sent at once. Also used as
    # ``bind_limit`` if that is 0 or None.
    row_limit = 1024
    # The total number of available bind variables a single statement
    # can use.
    bind_limit = None
    # The default max_allowed_packet in MySQL is 4MB,
    # so the data, including encoding and the rest of the query structure,
    # must be less than that (unless we dynamically query to find out
    # what value we have)
    size_limit = parse_byte_size('2 MB')
    delete_placeholder = '%s'
    insert_placeholder = '%s'
    # For testing, force the delete order to be deterministic
    # when multiple columns are involved
    sorted_deletes = False

    def __init__(self, cursor, row_limit=None,
                 delete_placeholder=None,
                 insert_placeholder=None,
                 bind_limit=None):
        self.cursor = cursor
        if delete_placeholder is not None:
            self.delete_placeholder = delete_placeholder
        if insert_placeholder is not None:
            self.insert_placeholder = insert_placeholder
        if row_limit is not None:
            self.row_limit = row_limit
        if bind_limit is not None:
            self.bind_limit = bind_limit

        # These are cumulative
        self.total_rows_inserted = 0
        self.total_size_inserted = 0
        self.total_rows_deleted = 0

        # These all get reset at each flush()
        self.rows_added = 0
        self.size_added = 0
        self.bind_params_added = 0

        self.deletes = defaultdict(set)   # {(table, columns_tuple): set([(column_value,)])}
        self.inserts = defaultdict(dict)  # {(command, header, row_schema, suffix): {rowkey: [row]}}

    def __repr__(self):
        return "<%s at %x tins=%d tdel=%d prows=%d>" % (
            self.__class__.__name__,
            id(self),
            self.total_rows_inserted,
            self.total_rows_deleted,
            self.rows_added,
        )

    def _flush_if_needed(self):
        if self.rows_added >= self.row_limit:
            return self.flush()
        if self.bind_limit and self.bind_params_added >= self.bind_limit:
            return self.flush()
        if self.size_added >= self.size_limit:
            return self.flush()

    def _flush_if_would_exceed_bind(self, addition):
        # The bind limit is a hard limit we cannot exceed.
        # If adding *addition* params would cause us to exceed,
        # flush now.
        if self.bind_limit and self.bind_params_added + addition >= self.bind_limit:
            self.flush()
            return True

    def delete_from(self, table, **kw):
        if not kw:
            raise AssertionError("Need at least one column value")
        columns = tuple(sorted(kw))
        key = (table, columns)
        row = tuple(kw[column] for column in columns)
        bind_params_added = len(row) if key not in self.deletes[key] else 0
        self._flush_if_would_exceed_bind(bind_params_added)

        self.deletes[key].add(row)
        self.rows_added += 1
        self.bind_params_added += bind_params_added
        self._flush_if_needed()

    def insert_into(self, header, row_schema, row, rowkey, size,
                    command='INSERT', suffix=''):
        key = (command, header, row_schema, suffix)

        bind_params_added = len(row) if rowkey not in self.inserts[key] else 0
        self._flush_if_would_exceed_bind(bind_params_added)

        # If we flushed, self.inserts has started all over.
        self.inserts[key][rowkey] = row  # note that this may replace a row

        self.rows_added += 1
        self.bind_params_added += bind_params_added
        self.size_added += size
        self._flush_if_needed()

    def row_schema_of_length(self, param_count):
        # Use as the *row_schema* parameter to insert_into
        key = (self.insert_placeholder, param_count)
        try:
            return _row_schemas[key]
        except KeyError:
            p = [self.insert_placeholder] * param_count
            p = ', '.join(p)
            _row_schemas[key] = p
            return p

    def select_from(self, columns, table, suffix='', timeout=None, **kw):
        """
        Handles a query of the ``WHERE col IN (?, ?,)`` type::

        ``SELECT columns FROM table WHERE col IN (?, ...)``

        The keyword arguments should be of length 1, containing an
        iterable of the values to check: ``col=(1, 2)`` or in the
        dynamic case ``**{indirect_var: [1, 2]}``::

            batcher.select_from(('zoid', 'tid',), 'object_state',
                                zoid=oids)

        The number of batches needed is determined by the length of
        the iterator divided by this object's ``bind_limit``,
        or, if that's not set, by the ``row_limit``.

        Returns a iterator of matching rows. Matching rows are delivered
        incrementally, so some number of rows may be delivered and then
        an exception is raised.

        :keyword float timeout: If given, provides a number of seconds
           that is the approximate maximum amount of time this method will
           be allowed to take.
        :raises AggregateOperationTimeoutError: If *timeout* is given,
           and the cumulative time taken to query and process
           some subset of batches exceeds *timeout*. This is checked
           after each individual batch.
        """
        assert len(kw) == 1
        filter_column, filter_values = kw.popitem()
        filter_values = iter(filter_values)

        command = 'SELECT %s' % (','.join(columns),)

        chunk_size = self.bind_limit or self.row_limit
        chunk_size -= 1

        begin = self.perf_counter() if timeout else None

        for head in filter_values:
            filter_subset = list(itertools.islice(filter_values, chunk_size))
            filter_subset.append(head)

            descriptor = [[(table, (filter_column,)), filter_subset]]

            self._do_batch(command, descriptor, rows_need_flattened=False, suffix=suffix)

            for row in self.cursor.fetchall():
                yield row

            if timeout and self.perf_counter() - begin >= timeout:
                # TODO: It'd be nice not to do this if we had no more
                # batches to do.
                raise AggregateOperationTimeoutError

    def flush(self):
        if self.deletes:
            self.total_rows_deleted += self._do_deletes()
            self.deletes.clear()
        if self.inserts:
            self.total_rows_inserted += self._do_inserts()
            self.inserts.clear()
        self.total_size_inserted += self.size_added

        self.rows_added = 0
        self.size_added = 0
        self.bind_params_added = 0

    def _do_deletes(self):
        return self._do_batch('DELETE', sorted(iteritems(self.deletes)))

    def _do_batch(self, command, descriptors, rows_need_flattened=True, suffix=''):
        count = 0
        for (table, columns), rows in descriptors:
            count += len(rows)
            # Be careful not to use % formatting on a string already
            # containing `delete_placeholder`
            these_params_need_flattened = rows_need_flattened
            if len(columns) == 1:
                stmt, params, these_params_need_flattened = self._make_single_column_query(
                    command, table,
                    columns[0], rows, rows_need_flattened
                )
            else:
                row_template = " AND ".join(
                    ("%s=" % (column,)) + self.delete_placeholder
                    for column in columns
                )
                row_template = "(%s)" % (row_template,)
                rows_template = [row_template] * len(rows)
                stmt = "%s FROM %s WHERE %s" % (
                    command,
                    table, " OR ".join(rows_template))
                params = rows

            if these_params_need_flattened:
                params = self._flatten_params(params)
            stmt += suffix
            self.cursor.execute(stmt, params)

        return count

    def _flatten_params(self, params):
        params = sorted(params) if self.sorted_deletes else params
        params = list(itertools.chain.from_iterable(params))
        return params

    def _make_placeholder_list_of_length(self, count):
        return ','.join([self.delete_placeholder] * count)

    def _make_single_column_query(self, command, table,
                                  filter_column, filter_value,
                                  rows_need_flattened):
        placeholder_str = self._make_placeholder_list_of_length(len(filter_value))
        stmt = "%s FROM %s WHERE %s IN (%s)" % (
            command, table, filter_column, placeholder_str
        )
        return stmt, filter_value, rows_need_flattened

    def _do_inserts(self):
        count = 0
        # In the case that we have only a single table we're inserting into,
        # and thus common key values, we don't need to sort or iterate.
        # If we have multiple tables, we never want to sort by the rows too,
        # that doesn't make any sense.
        if len(self.inserts) == 1:
            items = [self.inserts.popitem()]
        else:
            items = sorted(iteritems(self.inserts), key=lambda i: i[0])
        for (command, header, row_schema, suffix), rows in items:
            # Batched inserts
            rows = list(rows.values())
            count += len(rows)
            value_template = "(%s)" % row_schema
            values_template = [value_template] * len(rows)
            params = list(itertools.chain.from_iterable(rows))

            stmt = "%s INTO %s VALUES\n%s\n%s" % (
                command, header, ',\n'.join(values_template), suffix)
            # e.g.,
            # INSERT INTO table(c1, c2)
            # VALUES (%s, %s), (%s, %s), (%s, %s)
            # <suffix>
            __traceback_info__ = stmt
            self.cursor.execute(stmt, params)
        return count



##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
TPC protocol state management.

The various states in which a storage instance can find itself during
two-phase commit are complicated. This package presents a set of
objects that encapsulate various possibilities. In this way we can
test independent states...independently, and the state transitions are
explicit.

"""
from __future__ import absolute_import
from __future__ import print_function

import logging
import os

from transaction.interfaces import NoTransaction
from transaction._transaction import rm_key
from transaction import get as get_thread_local_transaction

from zope.interface import implementer

from ZODB.POSException import ReadOnlyError
from ZODB.POSException import StorageTransactionError

from ..interfaces import ITPCStateNotInTransaction
from ..interfaces import ITPCStateDatabaseAvailable
from ...adapters.connections import ClosedConnection
from ..._util import Lazy as BaseLazy

from .temporary_storage import TemporaryStorage

log = logging.getLogger("relstorage")

_CLOSED_CONNECTION = ClosedConnection()

#: Set the ``RELSTORAGE_LOCK_EARLY`` environment variable if you
#: experience deadlocks or failures to commit (``tpc_finish``). This
#: will cause the commit lock to be taken as part of ``tpc_vote``
#: (similar to RelStorage 2.x) instead of deferring it until
#: ``tpc_finish``.
#:
#: If this is necessary, this is probably a bug in RelStorage; please report
#: it.
LOCK_EARLY = os.environ.get('RELSTORAGE_LOCK_EARLY')


class _LazyResource(BaseLazy):

    # If not None, a callable ``(storage, resource, force)``
    # that aborts the *resource*, possibly forcefully (*force*).
    # The return value will be the new value in the object
    # instance.
    abort_function = None
    # If not None, a callable ``(storage, resource)`` to clean up
    # any use of the *resource* after success.
    release_function = None

    def _stored_value_for_name_in_inst(self, value, name, inst):
        # type: (Any, str, SharedTPCState) -> None
        if name == 'store_connection':
            # Try to do this first
            inst._used_resources.insert(0, self)
        else:
            inst._used_resources.append(self)

    def aborter(self, func):
        self.abort_function = func
        return self

    def releaser(self, func):
        self.release_function = func
        return self


class SharedTPCState(object):
    """
    Contains attributes marking resources that *might* be used during the commit
    process. If any of them are, then the `abort` method takes care of cleaning them up.

    Accessing a resource implicitly begins it, if needed.
    """

    # pylint:disable=method-hidden

    prepared_txn = None
    transaction = None
    not_in_transaction_state = None
    read_only = False # Or we wouldn't allocate this object.

    def __init__(self, initial_state, storage, transaction):
        self.initial_state = initial_state
        self._storage = storage
        self.transaction = transaction
        self._used_resources = []

    @_LazyResource
    def store_connection(self):
        return self._storage._store_connection_pool.borrow()

    @store_connection.aborter
    def store_connection(self, storage, store_connection, force):
        try:
            adapter = storage._adapter
            if store_connection:
                # It's possible that this connection/cursor was
                # already closed if an error happened (which would
                # release the locks). Don't try to re-open it.
                adapter.locker.release_commit_lock(store_connection.cursor)

                # Though, this might re-open it.
                adapter.txncontrol.abort(
                    store_connection,
                    self.prepared_txn)

                if force:
                    store_connection.drop()
        finally:
            storage._store_connection_pool.replace(store_connection)
        return _CLOSED_CONNECTION

    @store_connection.releaser
    def store_connection(self, storage, store_connection):
        storage._store_connection_pool.replace(store_connection)
        return _CLOSED_CONNECTION

    @_LazyResource
    def load_connection(self):
        return self._storage._load_connection

    @load_connection.aborter
    def load_connection(self, _storage, load_connection, force):
        if force:
            load_connection.drop()
        else:
            load_connection.rollback_quietly()
        load_connection.exit_critical_phase()
        return _CLOSED_CONNECTION

    @load_connection.releaser
    def load_connection(self, _storage, load_connection):
        load_connection.rollback_quietly()
        load_connection.exit_critical_phase()
        return _CLOSED_CONNECTION

    @_LazyResource
    def blobhelper(self):
        blobhelper = self._storage.blobhelper
        blobhelper.begin()
        return blobhelper

    @blobhelper.aborter
    def blobhelper(self, _storage, blobhelper, _force):
        blobhelper.abort()

    @blobhelper.releaser
    def blobhelper(self, _storage, blobhelper):
        blobhelper.clear_temp()

    def has_blobs(self):
        # pylint:disable=no-member
        return (
            'blobhelper' in self.__dict__
            and self.blobhelper is not None
            and self.blobhelper.txn_has_blobs
        )

    @BaseLazy
    def cache(self):
        return self._storage._cache

    @BaseLazy
    def adapter(self):
        return self._storage._adapter

    @_LazyResource
    def temp_storage(self):
        return TemporaryStorage()

    @temp_storage.aborter
    def temp_storage(self, _storage, temp_storage, _force):
        temp_storage.close()

    @temp_storage.releaser
    def temp_storage(self, _storage, temp_storage):
        temp_storage.close()

    def has_temp_data(self):
        return 'temp_storage' in self.__dict__ and self.temp_storage

    def __cleanup(self, method_name, method_args):
        storage = self._storage
        resources = self._used_resources
        self._used_resources = () # No more opening resources.

        exceptions = []

        for resource in resources:
            assert resource.__name__ in vars(self)

            cleaner = getattr(resource, method_name)
            if not cleaner:
                setattr(self, resource.__name__, None)
                continue

            value = getattr(self, resource.__name__)
            new_value = None
            try:
                new_value = cleaner(self, storage, value, *method_args)
            except Exception as ex: # pylint:disable=broad-except
                exceptions.append(ex)
            setattr(self, resource.__name__, new_value)

        if exceptions: # pragma: no cover
            raise Exception("Failed to close one or more resources: %s" % (exceptions,))

    def abort(self, force=False):
        self.__cleanup('abort_function', (force,))

    def release(self):
        self.__cleanup('release_function', ())


@implementer(ITPCStateDatabaseAvailable)
class AbstractTPCStateDatabaseAvailable(object):

    __slots__ = (
        'shared_state',
    )

    # - store
    # - restore/restoreBlob
    # - deleteObject
    # - undo

    # should raise ReadOnlyError if the storage is read only.

    # - tpc_vote should raise StorageTransactionError

    # Because entering tpc_begin wasn't allowed if the storage was
    # read only, this needs to happen in the "not in transaction"
    # state.

    def __init__(self, shared_state):
        self.shared_state = shared_state # type: SharedTPCState

    @property
    def transaction(self):
        return self.shared_state.transaction

    @property
    def initial_state(self):
        return self.shared_state.initial_state

    @property
    def store_connection(self):
        return self.shared_state.store_connection

    def __repr__(self):
        result = "<%s at 0x%x stored_count=%s %s" % (
            type(self).__name__,
            id(self),
            len(getattr(self, 'temp_storage', ()) or ()),
            self._tpc_state_transaction_data(),
        )

        extra = self._tpc_state_extra_repr_info()
        for k, v in extra.items():
            result += ' %s=%r' % (k, v)
        result += '>'
        return result

    def _tpc_state_extra_repr_info(self):
        return {}

    def _tpc_state_transaction_data(self):
        # Grovels around in the transaction object and tries to find interesting
        # things to include.

        # The ZODB Connection passes us an internal TransactionMetaData
        # object; the real transaction object stores a reference to that in its data,
        # keyed off the connection.
        # We may or may not be able to get the real transaction using transaction.get(),
        # depending on if we are using the global (thread local) transaction manager or not.
        try:
            global_tx = get_thread_local_transaction()
        except NoTransaction:
            # It's in explicit mode and we're not using it.
            return "<no global transaction> tx=%r" % (self.transaction,)

        tx_data = getattr(global_tx, '_data', None)
        if not tx_data:
            # No data stored on the transaction (or the implementation changed!)
            return "<no transaction data> tx=%r" % (self.transaction,)

        for v in tx_data.values():
            if v is self.transaction:
                # Yes, we found the metadata that ZODB uses, so we are
                # joined to this transaction.
                break
        else:
            return "<no transaction meta %r> tx=%r" % (tx_data, self.transaction,)

        resources = sorted(global_tx._resources, key=rm_key)
        return "transaction=%r resources=%r" % (global_tx, resources)

    def tpc_finish(self, storage, transaction, f=None): # pylint:disable=unused-argument
        # For the sake of some ZODB tests, we need to implement this everywhere,
        # even if it's not actually usable, and the first thing it needs to
        # do is check the transaction.
        if transaction is not self.transaction:
            raise StorageTransactionError('tpc_finish called with wrong transaction')
        raise NotImplementedError("tpc_finish not allowed in this state.")

    def tpc_begin(self, _storage, transaction):
        # Ditto as for tpc_finish
        raise StorageTransactionError('tpc_begin not allowed in this state', type(self))

    def tpc_abort(self, transaction, force=False):
        if not force:
            if transaction is not self.transaction:
                return self

        self.shared_state.abort(force)
        return self.initial_state

    def no_longer_stale(self):
        return self

    def stale(self, e):
        return Stale(self, e)

    def close(self):
        if self.shared_state is not None:
            self.tpc_abort(None, True)
            self.shared_state = None


@implementer(ITPCStateNotInTransaction)
class NotInTransaction(object):
    # The default state, when the storage is not attached to a
    # transaction.

    __slots__ = (
        'last_committed_tid_int',
        'read_only',
        'begin_factory',
    )

    transaction = None

    def __init__(self, begin_factory, read_only, committed_tid_int=0):
        self.begin_factory = begin_factory
        self.read_only = read_only
        self.last_committed_tid_int = committed_tid_int

    def with_committed_tid_int(self, committed_tid_int):
        return NotInTransaction(
            self.begin_factory,
            self.read_only,
            committed_tid_int
        )

    def tpc_abort(self, *args, **kwargs): # pylint:disable=arguments-differ,unused-argument,signature-differs
        # Nothing to do
        return self

    def _no_transaction(self, *args, **kwargs):
        raise StorageTransactionError("No transaction in progress")

    tpc_finish = tpc_vote = _no_transaction
    checkCurrentSerialInTransaction = _no_transaction

    def store(self, *_args, **_kwargs):
        if self.read_only:
            raise ReadOnlyError()
        self._no_transaction()

    restore = deleteObject = undo = restoreBlob = store

    def tpc_begin(self,  storage, transaction): # XXX: Signature needs to change.
        if self.read_only:
            raise ReadOnlyError()
        if transaction is self.transaction: # Also handles None.
            raise StorageTransactionError("Duplicate tpc_begin calls for same transaction.")
        state = SharedTPCState(self, storage, transaction)
        try:
            return self.begin_factory(state)
        except:
            state.abort()
            raise

    @property
    def initial_state(self):
        return self

    # This object appears to be false.
    def __bool__(self):
        return False
    __nonzero__ = __bool__

    def close(self):
        pass


@implementer(ITPCStateNotInTransaction)
class Stale(object):
    """
    An error that lets us know we are stale
    was encountered.

    Just about all accesses to this object result in
    re-raising that error.
    """

    transaction = None
    last_committed_tid_int = 0

    def __init__(self, previous_state, stale_error):
        self.previous_state = previous_state
        self.stale_error = stale_error

    def _stale(self, *args, **kwargs):
        raise self.stale_error

    store = restore = checkCurrentSerialInTransaction = _stale
    undo = deleteObject = restoreBlob = _stale
    tpc_begin = tpc_finish = tpc_vote = _stale

    def tpc_abort(self, *args, **kwargs):
        return self.previous_state.tpc_abort(*args, **kwargs)

    @property
    def initial_state(self):
        return self.previous_state.initial_state

    def no_longer_stale(self):
        return self.previous_state

    def stale(self, _e):
        return self

    def __bool__(self):
        return False
    __nonzero__ = __bool__



##############################################################################
#
# Copyright (c) 2017 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest

from relstorage.adapters._util import noop_when_history_free

from ..mover import metricmethod_sampled


class TestNoOp(unittest.TestCase):


    def test_history_free(self):

        class Thing(object):

            keep_history = False

            @noop_when_history_free
            @metricmethod_sampled
            def thing(self, arg1, arg2=False):
                "Docs"
                assert arg1
                assert arg2


        thing = Thing()
        # Before calling it has a wrapper
        thing_thing = thing.thing
        self.assertTrue(hasattr(thing_thing, '__wrapped__'))
        self.assertEqual(Thing.thing.__doc__, "Docs")
        self.assertEqual(Thing.thing.__doc__, thing_thing.__doc__)
        self.assertEqual(Thing.thing.__name__, 'thing')
        self.assertEqual(Thing.thing.__name__, thing_thing.__name__)

        # call with false args to be sure doesn't run
        thing.thing(0) # one argument
        thing.thing(0, arg2=0) # kwarg

        # It still has a wrapper, but it's not the original
        self.assertTrue(hasattr(thing.thing, '__wrapped__'))
        self.assertIsNot(thing.thing, thing_thing)
        self.assertEqual(Thing.thing.__doc__, thing.thing.__doc__)
        self.assertEqual(Thing.thing.__name__, 'thing')
        self.assertEqual(Thing.thing.__name__, thing.thing.__name__)

    def test_keep_history(self):

        class Thing(object):

            keep_history = True
            ran = False

            @noop_when_history_free
            def thing(self, arg1, arg2=False):
                "Docs"
                assert arg1
                assert arg2
                self.ran = True
                return arg1


        thing = Thing()
        # Before calling it has a wrapper
        thing_thing = thing.thing
        self.assertTrue(hasattr(thing_thing, '__wrapped__'))
        self.assertEqual(Thing.thing.__doc__, "Docs")
        self.assertEqual(Thing.thing.__doc__, thing_thing.__doc__)
        self.assertEqual(Thing.thing.__name__, 'thing')
        self.assertEqual(Thing.thing.__name__, thing_thing.__name__)

        # call with false args to be sure does run
        self.assertRaises(AssertionError, thing.thing, 0) # one argument
        thing.thing(1, arg2=1)
        self.assertRaises(AssertionError, thing.thing, 1, arg2=0)

        # It has access to self
        self.assertTrue(thing.ran)

        # It returns a value...
        self.assertIs(self, thing.thing(self, 1))
        # ...even the first time its called
        self.assertIs(self, Thing().thing(self, 1))

        # It still has a wrapper, but it's not the original
        self.assertTrue(hasattr(thing.thing, '__wrapped__'))
        self.assertIsNot(thing.thing, thing_thing)
        self.assertEqual(Thing.thing.__doc__, thing.thing.__doc__)
        self.assertEqual(Thing.thing.__name__, 'thing')
        self.assertEqual(Thing.thing.__name__, thing.thing.__name__)

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestNoOp))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Batch table row insert/delete support.
"""

from __future__ import absolute_import

import re

from relstorage._compat import list_values

from ..batch import RowBatcher

oracle_rowvar_re = re.compile(r":([a-zA-Z0-9_]+)")

class OracleRowBatcher(RowBatcher):
    """
    Oracle-specific row batcher.

    Expects :name parameters and a dictionary for each row.
    """

    insert_placeholder = delete_placeholder = ':1'

    # ORA-01795: maximum number of expressions in a list is 1000
    row_limit = 999

    def __init__(self, cursor, inputsizes, row_limit=None):
        super(OracleRowBatcher, self).__init__(cursor, row_limit)
        self.inputsizes = inputsizes
        self.array_ops = {}  # {(operation, row_schema): {rowkey: [row]}}

    def _do_inserts(self):
        # pylint:disable=too-many-locals
        def replace_var(match):
            name = match.group(1)
            new_name = '%s_%d' % (name, rownum) # pylint:disable=undefined-loop-variable
            if name in self.inputsizes:
                stmt_inputsizes[new_name] = self.inputsizes[name]
            params[new_name] = row[name]
            return ':%s' % new_name

        items = sorted(self.inserts.items())
        for (_command, header, row_schema, _), rows in items:
            stmt_inputsizes = {}

            if len(rows) == 1:
                # use the single insert syntax
                row = list_values(rows)[0]
                stmt = "INSERT INTO %s VALUES (%s)" % (header, row_schema)
                for name in self.inputsizes:
                    if name in row:
                        stmt_inputsizes[name] = self.inputsizes[name]
                if stmt_inputsizes:
                    self.cursor.setinputsizes(**stmt_inputsizes)
                self.cursor.execute(stmt, row)

            else:
                # use the multi-insert syntax
                parts = []
                params = {}
                for rownum, row in enumerate(rows.values()):
                    mod_row = oracle_rowvar_re.sub(replace_var, row_schema)
                    parts.append("INTO %s VALUES (%s)" % (header, mod_row))


                stmt = "INSERT ALL\n%s\nSELECT * FROM DUAL" % '\n'.join(parts)
                if stmt_inputsizes:
                    self.cursor.setinputsizes(**stmt_inputsizes)
                self.cursor.execute(stmt, params)

    def add_array_op(self, operation, row_schema, row, rowkey, size):
        key = (operation, row_schema)
        rows = self.array_ops.get(key)
        if rows is None:
            self.array_ops[key] = rows = {}
        rows[rowkey] = row  # note that this may replace a row
        self.rows_added += 1
        self.size_added += size
        if (self.rows_added >= self.row_limit
                or self.size_added >= self.size_limit):
            self.flush()

    def flush(self):
        if self.deletes:
            self._do_deletes()
            self.deletes.clear()
        if self.inserts:
            self._do_inserts()
            self.inserts.clear()
        if self.array_ops:
            self._do_array_ops()
            self.array_ops.clear()
        self.rows_added = 0
        self.size_added = 0

    def _do_array_ops(self):
        items = sorted(self.array_ops.items())
        for (operation, row_schema), rows in items:
            r = list_values(rows)
            params = []
            datatypes = [self.inputsizes[name] for name in row_schema.split()]
            for i, column in enumerate(zip(*r)):
                params.append(self.cursor.arrayvar(datatypes[i], list(column)))
            self.cursor.execute(operation, tuple(params))

    # TODO: Can we override _make_single_column_query() to pass
    # an array to the database for use in a sql statement?
    # We'd need the TABLE operator, probably. How far back is that supported?

    def _make_placeholder_list_of_length(self, count):
        return ','.join((':%d' % i for i in range(count)))



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import print_function

import functools
import gc
import os
import tempfile
import unittest
from contextlib import contextmanager

import transaction
from zc.zlibstorage import ZlibStorage
from ZODB.DB import DB
from ZODB.FileStorage import FileStorage

from relstorage.zodbconvert import main


def skipIfZapNotSupportedByDest(func):
    @functools.wraps(func)
    def test(self):
        if not self.zap_supported_by_dest:
            raise unittest.SkipTest("zap_all not supported")
        func(self)
    return test

class AbstractZODBConvertBase(unittest.TestCase):
    cfgfile = None

    # Set to True in a subclass if the destination can be zapped
    zap_supported_by_dest = False

    def setUp(self):
        super(AbstractZODBConvertBase, self).setUp()
        self._to_close = []

    def tearDown(self):
        for i in self._to_close:
            i.close()
        self._to_close = []
        # XXX: On PyPy with psycopg2cffi, running these two tests will
        # result in a hang:
        # HPPostgreSQLDestZODBConvertTests.test_clear_empty_dest
        # HPPostgreSQLDestZODBConvertTests.test_clear_full_dest
        # test_clear_full_dest will hang in the zodbconvert call to
        # zap_all(), in the C code of the PG driver. Presumably some
        # connection with some lock got left open and was preventing
        # the TRUNCATE statements from taking out a lock. The same
        # tests do not hang with psycopg2cffi on C Python. Manually
        # running the gc (twice!) here fixes the issue. Note that this
        # only started when we wrapped the destination storage in
        # ZlibStorage (which copies methods into its own dict) so
        # there's something weird going on with the GC. Seen in PyPy
        # 2.5.0 and 5.3.
        gc.collect()
        gc.collect()
        super(AbstractZODBConvertBase, self).tearDown()

    def _closing(self, thing):
        self._to_close.append(thing)
        return thing

    def _create_src_storage(self):
        raise NotImplementedError()

    def _create_dest_storage(self):
        raise NotImplementedError()

    def _create_src_db(self):
        return self._closing(DB(self._closing(self._create_src_storage())))

    def _create_dest_db(self):
        return self._closing(DB(self._closing(self._create_dest_storage())))

    @contextmanager
    def __conn(self, name):
        db = getattr(self, '_create_' + name + '_db')()
        conn = db.open()
        try:
            yield conn
        finally:
            conn.close()
            db.close()

    def _src_conn(self):
        return self.__conn('src')

    def _dest_conn(self):
        return self.__conn('dest')

    def __write_value_for_key_in_db(self, val, key, db_conn_func):
        with db_conn_func() as conn:
            conn.root()[key] = val
            transaction.commit()

    def _write_value_for_key_in_src(self, x, key='x'):
        self.__write_value_for_key_in_db(x, key, self._src_conn)

    def _write_value_for_key_in_dest(self, x, key='x'):
        self.__write_value_for_key_in_db(x, key, self._dest_conn)

    def _check_value_of_key_in_dest(self, x, key='x'):
        with self._dest_conn() as conn2:
            db_x = conn2.root().get(key)
            self.assertEqual(db_x, x)

    def test_convert(self):
        self._write_value_for_key_in_src(10)
        main(['', self.cfgfile])
        self._check_value_of_key_in_dest(10)


    def test_dry_run(self):
        self._write_value_for_key_in_src(10)
        main(['', '--dry-run', self.cfgfile])
        self._check_value_of_key_in_dest(None)

    def test_incremental(self):
        self._write_value_for_key_in_src(10)
        main(['', self.cfgfile])
        self._check_value_of_key_in_dest(10)

        self._write_value_for_key_in_src("hi")
        main(['', '--incremental', self.cfgfile])
        self._check_value_of_key_in_dest("hi")

    def test_incremental_empty_src_dest(self):
        # Should work and not raise a POSKeyError
        main(['', '--incremental', self.cfgfile])
        self._check_value_of_key_in_dest(None)

    @skipIfZapNotSupportedByDest
    def test_clear_empty_dest(self):
        x = 10
        self._write_value_for_key_in_src(x)
        main(['', '--clear', self.cfgfile])
        self._check_value_of_key_in_dest(x)

    @skipIfZapNotSupportedByDest
    def test_clear_full_dest(self):
        self._write_value_for_key_in_dest(999)
        self._write_value_for_key_in_dest(666, key='y')
        self._write_value_for_key_in_dest(8675309, key='z')

        self._write_value_for_key_in_src(1, key='x')
        self._write_value_for_key_in_src(2, key='y')
        # omit z

        main(['', '--clear', self.cfgfile])

        self._check_value_of_key_in_dest(1, key='x')
        self._check_value_of_key_in_dest(2, key='y')
        self._check_value_of_key_in_dest(None, key='z')

    def test_no_overwrite(self):
        db = self._create_src_db() # create the root object
        db.close()
        db = self._create_dest_db() # create the root object
        db.close()
        self.assertRaises(SystemExit, main, ['', self.cfgfile])

class FSZODBConvertTests(AbstractZODBConvertBase):

    def setUp(self):
        super(FSZODBConvertTests, self).setUp()

        fd, self.srcfile = tempfile.mkstemp()
        os.close(fd)
        os.remove(self.srcfile)

        fd, self.destfile = tempfile.mkstemp()
        os.close(fd)
        os.remove(self.destfile)

        cfg = self._cfg_header() + self._cfg_source() + self._cfg_dest()
        self._write_cfg(cfg)

    def _cfg_header(self):
        return ""

    def _cfg_source(self):
        return """
        <filestorage source>
            path %s
        </filestorage>
        """ % self.srcfile

    def _cfg_dest(self):
        return """
        <filestorage destination>
            path %s
        </filestorage>
        """ % self.destfile

    def _write_cfg(self, cfg):
        fd, self.cfgfile = tempfile.mkstemp()
        os.write(fd, cfg.encode('ascii'))
        os.close(fd)

    def tearDown(self):
        if os.path.exists(self.destfile):
            os.remove(self.destfile)
        if os.path.exists(self.srcfile):
            os.remove(self.srcfile)
        if os.path.exists(self.cfgfile):
            os.remove(self.cfgfile)
        super(FSZODBConvertTests, self).tearDown()

    def _create_src_storage(self):
        return FileStorage(self.srcfile)

    def _create_dest_storage(self):
        return FileStorage(self.destfile)

    def test_storage_has_data(self):
        from relstorage.zodbconvert import storage_has_data
        src = FileStorage(self.srcfile, create=True)
        self.assertFalse(storage_has_data(src))
        db = DB(src)  # add the root object
        db.close()
        self.assertTrue(storage_has_data(src))

class ZlibWrappedZODBConvertTests(FSZODBConvertTests):

    def _cfg_header(self):
        return "%import zc.zlibstorage\n"

    def _cfg_source(self):
        return ("\n<zlibstorage source>"
                + super(ZlibWrappedZODBConvertTests, self)._cfg_source()
                + "</zlibstorage>")

    def _cfg_dest(self):
        return ("\n<zlibstorage destination>"
                + super(ZlibWrappedZODBConvertTests, self)._cfg_dest()
                + "</zlibstorage>")

    def _create_src_storage(self):
        return ZlibStorage(super(ZlibWrappedZODBConvertTests, self)._create_src_storage())

    def _create_dest_storage(self):
        return ZlibStorage(super(ZlibWrappedZODBConvertTests, self)._create_dest_storage())


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(FSZODBConvertTests))
    suite.addTest(unittest.makeSuite(ZlibWrappedZODBConvertTests))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""IObjectMover implementation.
"""
from __future__ import absolute_import

import os
import sys

from zope.interface import implementer

from ..interfaces import IObjectMover
from ..mover import AbstractObjectMover
from ..mover import RowBatcherStoreTemps
from ..mover import metricmethod_sampled

class OracleRowBatcherStoreTemps(RowBatcherStoreTemps):

    def store_temp_into_batcher(self, batcher, oid, prev_tid, data):
        md5sum = self._compute_md5sum(data)

        size = len(data)
        if size <= 2000:
            # Send data inline for speed.  Oracle docs say maximum size
            # of a RAW is 2000 bytes.
            stmt = "BEGIN relstorage_op.store_temp(:1, :2, :3, :4); END;"
            batcher.add_array_op(
                stmt,
                'oid prev_tid md5sum rawdata',
                (oid, prev_tid, md5sum, data),
                rowkey=oid,
                size=size,
            )
        else:
            # Send data as a BLOB
            row = {
                'oid': oid,
                'prev_tid': prev_tid,
                'md5sum': md5sum,
                'blobdata': data,
            }
            batcher.insert_into(
                "temp_store (zoid, prev_tid, md5, state)",
                ":oid, :prev_tid, :md5sum, :blobdata",
                row,
                rowkey=oid,
                size=size,
            )


@implementer(IObjectMover)
class OracleObjectMover(OracleRowBatcherStoreTemps,
                        AbstractObjectMover):

    # This is assigned to by the adapter.
    inputsizes = None

    def __init__(self, *args, **kwargs):
        AbstractObjectMover.__init__(self, *args, **kwargs)
        OracleRowBatcherStoreTemps.__init__(
            self,
            self.keep_history,
            self.driver.Binary,
            batcher_factory=self.make_batcher)

    @metricmethod_sampled
    def get_object_tid_after(self, cursor, oid, tid):
        """Returns the tid of the next change after an object revision.

        Returns None if no later state exists.
        """
        stmt = """
        SELECT MIN(tid)
        FROM object_state
        WHERE zoid = :1
            AND tid > :2
        """
        cursor.execute(stmt, (oid, tid))
        rows = cursor.fetchall()
        if rows:
            # XXX: If we can use rowcount here, we can combine
            # with superclass.
            assert len(rows) == 1
            return rows[0][0]

    # no store connection initialization needed for Oracle
    def on_store_opened(self, cursor, restart=False):
        pass

    @metricmethod_sampled
    def restore(self, cursor, batcher, oid, tid, data):
        """Store an object directly, without conflict detection.

        Used for copying transactions into this database.
        """
        md5sum = self._compute_md5sum(data)

        size = len(data) if data is not None else 0

        if size <= 2000:
            # Send data inline for speed.  Oracle docs say maximum size
            # of a RAW is 2000 bytes.
            # XXX: Revisit this. cx_Oracle docs suggest that blobs up to 1GB can
            # be sent inline very fast.
            if self.keep_history:
                stmt = "BEGIN relstorage_op.restore(:1, :2, :3, :4); END;"
                batcher.add_array_op(
                    stmt,
                    'oid tid md5sum rawdata',
                    (oid, tid, md5sum, data),
                    rowkey=(oid, tid),
                    size=size,
                )
            else:
                stmt = "BEGIN relstorage_op.restore(:1, :2, :3); END;"
                batcher.add_array_op(
                    stmt,
                    'oid tid rawdata',
                    (oid, tid, data),
                    rowkey=(oid, tid),
                    size=size,
                )

        else:
            # Send as a BLOB
            if self.keep_history:
                row = {
                    'oid': oid,
                    'tid': tid,
                    'md5sum': md5sum,
                    'state_size': size,
                    'blobdata': data,
                }
                row_schema = """
                    :oid, :tid,
                    COALESCE((SELECT tid
                              FROM current_object
                              WHERE zoid = :oid), 0),
                    :md5sum, :state_size, :blobdata
                """
                batcher.insert_into(
                    "object_state (zoid, tid, prev_tid, md5, state_size, state)",
                    row_schema,
                    row,
                    rowkey=(oid, tid),
                    size=size,
                )
            else:
                batcher.delete_from('object_state', zoid=oid)
                if data:
                    row = {
                        'oid': oid,
                        'tid': tid,
                        'state_size': size,
                        'blobdata': data,
                    }
                    batcher.insert_into(
                        "object_state (zoid, tid, state_size, state)",
                        ":oid, :tid, :state_size, :blobdata",
                        row,
                        rowkey=oid,
                        size=size,
                    )

    @metricmethod_sampled
    def download_blob(self, cursor, oid, tid, filename):
        """Download a blob into a file."""
        stmt = """
        SELECT chunk
        FROM blob_chunk
        WHERE zoid = :1
            AND tid = :2
        ORDER BY chunk_num
        """

        f = None
        bytecount = 0
        # Current versions of cx_Oracle only support offsets up
        # to sys.maxint or 4GB, whichever comes first.
        maxsize = min(sys.maxsize, 1 << 32)
        oth = cursor.connection.outputtypehandler
        del cursor.connection.outputtypehandler
        try:
            cursor.execute(stmt, (oid, tid))
            while True:
                try:
                    blob, = cursor.fetchone()
                except TypeError:
                    # No more chunks.  Note: if there are no chunks at
                    # all, then this method should not write a file.
                    break

                if f is None:
                    f = open(filename, 'wb')
                # round off the chunk-size to be a multiple of the oracle
                # blob chunk size to maximize performance
                read_chunk_size = int(
                    max(
                        round(1.0 * self.blob_chunk_size / blob.getchunksize()),
                        1)
                    * blob.getchunksize())
                offset = 1 # Oracle still uses 1-based indexing.
                reader = iter(lambda: blob.read(offset, read_chunk_size), b'')
                for read_chunk in reader:
                    f.write(read_chunk)
                    bytecount += len(read_chunk)
                    offset += len(read_chunk)
                    if offset > maxsize:
                        # We have already read the maximum we can store
                        # so we can assume we are done. If we do not break
                        # off here, cx_Oracle will throw an overflow
                        # exception anyway.
                        break
        except:
            if f is not None:
                f.close()
                os.remove(filename)
            raise
        finally:
            cursor.connection.outputtypehandler = oth

        if f is not None:
            f.close()
        return bytecount


    @metricmethod_sampled
    def upload_blob(self, cursor, oid, tid, filename):
        """Upload a blob from a file.

        If serial is None, upload to the temporary table.
        """
        # pylint:disable=too-many-locals
        if tid is not None:
            if self.keep_history:
                delete_stmt = """
                DELETE FROM blob_chunk
                WHERE zoid = :1 AND tid = :2
                """
                cursor.execute(delete_stmt, (oid, tid))
            else:
                delete_stmt = "DELETE FROM blob_chunk WHERE zoid = :1"
                cursor.execute(delete_stmt, (oid,))

            use_tid = True
            insert_stmt = """
            INSERT INTO blob_chunk (zoid, tid, chunk_num, chunk)
            VALUES (:oid, :tid, :chunk_num, empty_blob())
            RETURNING chunk INTO :newblob
            """
        else:
            use_tid = False
            delete_stmt = "DELETE FROM temp_blob_chunk WHERE zoid = :1"
            cursor.execute(delete_stmt, (oid,))

            insert_stmt = """
            INSERT INTO temp_blob_chunk (zoid, chunk_num, chunk)
            VALUES (:oid, :chunk_num, empty_blob())
            RETURNING chunk INTO :newblob
            """

        f = open(filename, 'rb')
        blob_var = cursor.var(self.driver.BLOB)
        params = dict(oid=oid, chunk_num=0, newblob=blob_var)
        if use_tid:
            params['tid'] = tid
        cursor.execute(insert_stmt, params)
        blob, = blob_var.getvalue()
        offset = 1 # Oracle uses 1-based indexing
        try:
            with f:
                while 1:
                    data = f.read(blob.getchunksize())
                    if data:
                        blob.write(data, offset)
                    else:
                        break
                    offset += len(data)

        finally:
            if blob is not None and blob.isopen():
                blob.close()



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Locker implementations.
"""

from __future__ import absolute_import
from __future__ import print_function

from contextlib import contextmanager

from zope.interface import implementer

from ..interfaces import ILocker
from ..interfaces import UnableToAcquireCommitLockError
from ..interfaces import UnableToAcquirePackUndoLockError
from ..locker import AbstractLocker

class CommitLockQueryFailedError(UnableToAcquireCommitLockError):
    pass

_SET_TIMEOUT_STMT = 'SET SESSION innodb_lock_wait_timeout = %s'
# DEFAULT is a literal, not a param, so cursor.execute(stmt, ('DEFAULT',))
# does not work.
_SET_TIMEOUT_DEFAULT_STMT = _SET_TIMEOUT_STMT % ('DEFAULT',)

@contextmanager
def lock_timeout(cursor, timeout, restore_to=None):
    """
    ContextManager that sets the lock timeout to the given value,
    and returns it to the DEFAULT when done.

    If *timeout* is ``None``, makes no changes to the connection.
    """
    if timeout is not None: # 0 is valid
        # Min value of timeout is 1; a value less than that produces
        # a warning but gets truncated to 1
        timeout = timeout if timeout >= 1 else 1
        cursor.execute(_SET_TIMEOUT_STMT, (timeout,))
        try:
            yield
        finally:
            if restore_to is None:
                cursor.execute(_SET_TIMEOUT_DEFAULT_STMT)
            else:
                cursor.execute(_SET_TIMEOUT_STMT, (restore_to,))
    else:
        yield


@implementer(ILocker)
class MySQLLocker(AbstractLocker):
    """
    MySQL locks.

    .. rubric:: Commit and Object Locks

    Two types of locks are used. The ordinary commit lock and the
    object locks are standard InnoDB row-level locks; this brings the
    benefits of being lightweight and automatically being released if
    the transaction aborts or commits, plus instant deadlock
    detection. Prior to MySQL 8.0, these don't support ``NOWAIT``
    syntax, so we synthesize that by setting the session variable
    `innodb_lock_wait_timeout
    <https://dev.mysql.com/doc/refman/5.7/en/innodb-parameters.html#sysvar_innodb_lock_wait_timeout>`_.

    Note that this lock cannot be against the ``object_state`` or
    ``current_object`` tables: arbitrary rows in those tables may have
    been locked by other transactions, and we risk deadlock.

    Also note that by default, a lock timeout will only rollback the
    current *statement*, not the whole transaction, as in most
    databases (this doesn't apply to ``NOWAIT`` in MySQL 8); to
    release any locks taken earlier, we must explicitly rollback the
    transaction. Fortunately, a lock timeout only rolling back the
    single statement is exactly what we want to implement ``NOWAIT``
    on earlier databases. In contrast, a detected deadlock will
    actually rollback the entire transaction.

    The ``ensure_current`` argument is essentially ignored; the locks
    taken out by ``lock_current_objects`` take care of that.

    .. rubric:: Shared and Exclusive Locks Can Block Each Other On Unrelated Rows

    We use two lock classes for object locks: shared locks for
    readCurrent, and exclusive locks for modified objects.

    MySQL 5.7 and 8 handle this weird, though. If two transactions are
    at any level besides ``SERIALIZABLE``, and one locks the *odd*
    rows ``FOR UPDATE`` the other one blocks trying to lock the *even*
    rows ``FOR UPDATE`` *or* in shared mode, if they happened to use
    queries like ``WHERE (zoid % 2) = 1``. This is surprising. (It's
    not surprising in ``SERIALIZABLE``; MySQL's ``SERIALIZABLE`` is
    quite pessimistic.)

    This is because (quoting
    https://dev.mysql.com/doc/refman/5.7/en/innodb-locks-set.html)
    "``SELECT ... LOCK IN SHARE MODE`` sets shared next-key locks on
    all index records the search encounters." While "``SELECT ... FOR
    UPDATE`` sets an exclusive next-key lock on every record the
    search encounters. However, only an index record lock is required
    for statements that lock rows using a unique index to search for a
    unique row. For index records the search encounters, ``SELECT ...
    FOR UPDATE`` blocks other sessions from doing ``SELECT ... LOCK IN
    SHARE MODE`` or from reading in certain transaction isolation
    levels." The complex ``WHERE`` clause does range queries and
    traversal of the index such that it winds up locking many
    unexpected rows.

    The good news is that the query we actually use for locking,
    ``SELECT zoid FROM ... WHERE zoid in (SELECT zoid from
    temp_store)``, doesn't do a range scan. It first accessess the
    ``temp_store`` table and does a sort into a temporary table using
    the index; then it accesses ``object_state`` or ``current_object``
    using the ``eq_ref`` method and the PRIMARY key index in a nested
    loop (sadly all MySQL joins are nested loops). This locks only the
    actually required rows.

    We should probably add some optimizer hints to make absolutely
    sure of that.

    .. rubric:: Pack Locks

    The second type of lock, an advisory lock, is used for pack locks.
    This lock uses the `GET_LOCK
    <https://dev.mysql.com/doc/refman/5.7/en/locking-functions.html#function_get-lock>`_
    and ``RELEASE_LOCK`` functions. These locks persist for the
    duration of a session, and *must* be explicitly released. They do
    *not* participate in deadlock detection.

    Prior to MySQL 5.7.5, it is not possible to hold more than one
    advisory lock in a single session. In the past we used advisory
    locks for the commit lock, and that meant we had to use multiple
    sessions (connections) to be able to hold both the commit lock and
    the pack lock. Fortunately, that limitation has been lifted: we no
    longer support older versions of MySQL, and we don't need multiple
    advisory locks anyway.
    """

    # The old MySQL 5.7 syntax is the default
    _lock_share_clause = 'LOCK IN SHARE MODE'
    _lock_share_clause_nowait = 'LOCK IN SHARE MODE'

    def __init__(self, options, driver, batcher_factory, version_detector):
        super(MySQLLocker, self).__init__(options, driver, batcher_factory)
        assert self.supports_row_lock_nowait # Set by default in the class.
        self.supports_row_lock_nowait = None
        self.version_detector = version_detector

        # No good preparing this, mysql can't take parameters in EXECUTE,
        # they have to be user variables, which defeats most of the point
        # (Although in this case, because it's a static value, maybe not;
        # it could be set once and re-used.)
        self.set_timeout_stmt = _SET_TIMEOUT_STMT

    def on_store_opened(self, cursor, restart=False):
        super(MySQLLocker, self).on_store_opened(cursor, restart=restart)
        if restart:
            return

        # Setting state in an `on_store_opened` is not good. With pooling, it may never
        # get executed. We work around this in our adapter by being sure to only use one
        # connmanager and one locker instance, shared among all the RelStorage instances.
        if self.supports_row_lock_nowait is None:
            self.supports_row_lock_nowait = self.version_detector.supports_nowait(cursor)

            if self.supports_row_lock_nowait:
                self._lock_share_clause = 'FOR SHARE'
                self._lock_share_clause_nowait = 'FOR SHARE NOWAIT'
            else:
                assert self._lock_readCurrent_oids_for_share
                self._lock_readCurrent_oids_for_share = self.__lock_readCurrent_nowait

    def _on_store_opened_set_row_lock_timeout(self, cursor, restart=False):
        self._set_row_lock_timeout(cursor, self.commit_lock_timeout)

    def _set_row_lock_timeout(self, cursor, timeout):
        # Min value of timeout is 1; a value less than that produces
        # a warning.
        timeout = timeout if timeout >= 1 else 1
        cursor.execute(self.set_timeout_stmt, (timeout,))
        # It's INCREDIBLY important to fetch a row after we execute the SET statement;
        # otherwise, the binary drivers that use libmysqlclient tend to crash,
        # usually with a 'malloc: freeing not allocated data' or 'malloc:
        # corrupted data, written after free?' or something like that.
        cursor.fetchone()


    def __lock_readCurrent_nowait(self, cursor, current_oids, shared_locks_block):
        # For MySQL 5.7, we emulate NOWAIT by setting the lock timeout
        if shared_locks_block:
            return AbstractLocker._lock_readCurrent_oids_for_share(self, cursor, current_oids, True)

        with lock_timeout(cursor, 0, self.commit_lock_timeout):
            return AbstractLocker._lock_readCurrent_oids_for_share(self, cursor, current_oids,
                                                                   False)

    def release_commit_lock(self, cursor):
        "Auto-released by transaction end."

    def _get_commit_lock_debug_info(self, cursor, was_failure=False):
        cursor.execute('SELECT connection_id()')
        conn_id = str(cursor.fetchone()[0])

        try:
            # MySQL 8
            cursor.execute("""
            SELECT *
            FROM performance_schema.events_transactions_current AS parent
                INNER JOIN performance_schema.data_locks AS child
                INNER JOIN performance_schema.data_lock_waits dlw on (child.engine_lock_id
                    = dlw.blocking_engine_lock_id)
            WHERE
            parent.THREAD_ID = child.THREAD_ID
            AND parent.EVENT_ID < child.EVENT_ID
            AND (
            child.EVENT_ID <= parent.END_EVENT_ID
            OR parent.END_EVENT_ID IS NULL
            )""")
            return 'Connection: ' + conn_id + '\n' + self._rows_as_pretty_string(cursor)
        except self.driver.driver_module.Error:
            # MySQL 5, or no permissions
            try:
                cursor.execute("""
                SELECT * from information_schema.innodb_locks l
                INNER JOIN information_schema.INNODB_TRX x ON l.lock_trx_id = x.trx_id
                """)
                rows = self._rows_as_pretty_string(cursor)
            except self.driver.driver_module.Error:
                # MySQL 8, and we had no permissions.
                return 'Connection: ' + conn_id
            return 'Connection: ' + conn_id + '\n' + rows

    def hold_pack_lock(self, cursor):
        """Try to acquire the pack lock.

        Raise an exception if packing or undo is already in progress.
        """
        stmt = "SELECT GET_LOCK(CONCAT(DATABASE(), '.pack'), 0)"
        cursor.execute(stmt)
        res = cursor.fetchone()[0]
        if not res:
            raise UnableToAcquirePackUndoLockError('A pack or undo operation is in progress')

    def release_pack_lock(self, cursor):
        """Release the pack lock."""
        stmt = "SELECT RELEASE_LOCK(CONCAT(DATABASE(), '.pack'))"
        cursor.execute(stmt)
        rows = cursor.fetchall() # stay in sync
        assert rows



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""TransactionControl implementations"""

from __future__ import absolute_import

import logging

from ..txncontrol import GenericTransactionControl
from ..txncontrol import noop_when_history_free

log = logging.getLogger(__name__)


class OracleTransactionControl(GenericTransactionControl):

    def __init__(self, connmanager, poller, keep_history, Binary, twophase):
        GenericTransactionControl.__init__(self, connmanager, poller, keep_history, Binary)
        self.twophase = twophase

    def commit_phase1(self, store_connection, tid):
        """Begin a commit.  Returns the transaction name.

        The transaction name must not be None.

        This method should guarantee that commit_phase2() will succeed,
        meaning that if commit_phase2() would raise any error, the error
        should be raised in commit_phase1() instead.
        """
        if self.twophase:
            store_connection.connection.prepare()
        return '-'

    @noop_when_history_free
    def add_transaction(self, cursor, tid, username, description, extension,
                        packed=False):
        stmt = """
        INSERT INTO transaction
            (tid, packed, username, description, extension)
        VALUES (:1, :2, :3, :4, :5)
        """
        max_desc_len = 2000
        if len(description) > max_desc_len:
            log.warning('Trimming description of transaction %s '
                        'to %d characters', tid, max_desc_len)
            description = description[:max_desc_len]
        cursor.execute(stmt, (
            tid, packed and 'Y' or 'N', self.Binary(username),
            self.Binary(description), self.Binary(extension)))



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Abstractions for using particular RDBMS implementations.

The interfaces in this package define what operations RelStorage needs
to work with a particular database. Most modules then have an Abstract
(partial) implementation of one such interface. The central
``IRelStorageAdapter`` interface ties them all together.

To use a new RDBMS, create a new package named for it, containing a
module named like each module found here to store the
implementation-specific class that subclasses the abstract version
found here.

Decisions that need to be made include:

- How to allocate OIDs.
- How to handle locking (commit and pack)
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest


class ReplicaSelectorTests(unittest.TestCase):

    def setUp(self):
        import os
        import tempfile
        fd, self.fn = tempfile.mkstemp()
        os.write(fd,
                 b"# Replicas\n\nexample.com:1234\nlocalhost:4321\n"
                 b"\nlocalhost:9999\n")
        os.close(fd)

    def tearDown(self):
        import os
        os.remove(self.fn)

    def test__read_config_normal(self):
        from relstorage.adapters.replica import ReplicaSelector
        rs = ReplicaSelector(self.fn, 600.0)
        self.assertEqual(rs._replicas,
                         ['example.com:1234', 'localhost:4321', 'localhost:9999'])

    def test__read_config_empty(self):
        from relstorage.adapters.replica import ReplicaSelector
        open(self.fn, 'w').close()  # truncate the replica list file
        self.assertRaises(IndexError, ReplicaSelector, self.fn, 600.0)

    def test__is_config_modified(self):
        from relstorage.adapters.replica import ReplicaSelector
        import time
        rs = ReplicaSelector(self.fn, 600.0)
        self.assertEqual(rs._is_config_modified(), False)
        # change the file
        rs._config_modified = 0
        # don't check the file yet
        rs._config_checked = time.time() + 3600
        self.assertEqual(rs._is_config_modified(), False)
        # now check the file
        rs._config_checked = 0
        self.assertEqual(rs._is_config_modified(), True)

    def test__select(self):
        from relstorage.adapters.replica import ReplicaSelector
        rs = ReplicaSelector(self.fn, 600.0)
        rs._select(0)
        self.assertEqual(rs._current_replica, 'example.com:1234')
        self.assertEqual(rs._current_index, 0)
        self.assertEqual(rs._expiration, None)
        rs._select(1)
        self.assertEqual(rs._current_replica, 'localhost:4321')
        self.assertEqual(rs._current_index, 1)
        self.assertNotEqual(rs._expiration, None)

    def test_current(self):
        from relstorage.adapters.replica import ReplicaSelector
        rs = ReplicaSelector(self.fn, 600.0)
        self.assertEqual(rs.current(), 'example.com:1234')
        # change the file and get the new current replica
        f = open(self.fn, 'w')
        f.write('localhost\nalternate\n')
        f.close()
        rs._config_checked = 0
        rs._config_modified = 0
        self.assertEqual(rs.current(), 'localhost')
        # switch to the alternate
        rs._select(1)
        self.assertEqual(rs.current(), 'alternate')
        # expire the alternate
        rs._expiration = 0
        self.assertEqual(rs.current(), 'localhost')

    def test_next_iteration(self):
        from relstorage.adapters.replica import ReplicaSelector
        rs = ReplicaSelector(self.fn, 600.0)

        # test forward iteration
        self.assertEqual(rs.current(), 'example.com:1234')
        self.assertEqual(rs.next(), 'localhost:4321')
        self.assertEqual(rs.next(), 'localhost:9999')
        self.assertEqual(rs.next(), None)

        # test iteration that skips over the replica that failed
        self.assertEqual(rs.current(), 'example.com:1234')
        self.assertEqual(rs.next(), 'localhost:4321')
        self.assertEqual(rs.current(), 'localhost:4321')
        # next() after current() indicates the last replica failed
        self.assertEqual(rs.next(), 'example.com:1234')
        self.assertEqual(rs.next(), 'localhost:9999')
        self.assertEqual(rs.next(), None)

    def test_next_only_one_server(self):
        from relstorage.adapters.replica import ReplicaSelector
        f = open(self.fn, 'w')
        f.write('localhost\n')
        f.close()
        rs = ReplicaSelector(self.fn, 600.0)
        self.assertEqual(rs.current(), 'localhost')
        self.assertEqual(rs.next(), None)

    def test_next_with_new_conf(self):
        from relstorage.adapters.replica import ReplicaSelector
        rs = ReplicaSelector(self.fn, 600.0)
        self.assertEqual(rs.current(), 'example.com:1234')
        self.assertEqual(rs.next(), 'localhost:4321')
        # interrupt the iteration by changing the replica conf file
        f = open(self.fn, 'w')
        f.write('example.com:9999\n')
        f.close()
        rs._config_checked = 0
        rs._config_modified = 0
        self.assertEqual(rs.next(), 'example.com:9999')
        self.assertEqual(rs.next(), None)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(ReplicaSelectorTests))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Oracle adapter for RelStorage."""
from __future__ import absolute_import

import logging

from ..._compat import metricmethod
from ..connmanager import AbstractConnectionManager
from ..interfaces import ReplicaClosedException

log = logging.getLogger(__name__)

class CXOracleConnectionManager(AbstractConnectionManager):

    isolation_read_committed = "ISOLATION LEVEL READ COMMITTED"

    # Note: the READ ONLY mode should be sufficient according to the
    # Oracle documentation, which says: "All subsequent queries in that
    # transaction see only changes that were committed before the
    # transaction began."
    #
    # See: http://download.oracle.com/docs/cd/B19306_01/server.102/b14200
    #   /statements_10005.htm
    #
    # This would be great for performance if we could rely on it.
    # It's like serializable isolation but with less locking.
    #
    # However, in testing an Oracle 10g RAC environment with
    # RelStorage, Oracle frequently leaked subsequently committed
    # transactions into a read only transaction, suggesting that read
    # only in RAC actually has read committed isolation rather than
    # serializable isolation. Switching to serializable mode solved the
    # problem. Using a DSN that specifies a particular RAC node did
    # *not* solve the problem. It's likely that this is a bug in RAC,
    # but let's be on the safe side and have all Oracle users apply
    # serializable mode instead of read only mode, since serializable
    # is more explicit.
    #
    # If anyone wants to try read only mode anyway, change the
    # class variable below.
    #
    #isolation_read_only = "READ ONLY"

    isolation_read_only = "ISOLATION LEVEL SERIALIZABLE"

    def __init__(self, driver, user, password, dsn, twophase, options):
        self._user = user
        self._password = password
        self._dsn = dsn
        self._twophase = twophase
        self._db_connect = driver.connect
        super(CXOracleConnectionManager, self).__init__(options, driver)

    @metricmethod
    def open(self,
             isolation=None,
             read_only=False,
             deferrable=False,
             replica_selector=None,
             application_name=None,
             twophase=False,
             **kwargs):
        """Open a database connection and return (conn, cursor)."""
        # pylint:disable=arguments-differ
        if isolation is None:
            isolation = 'ISOLATION LEVEL READ COMMITTED'

        if replica_selector is None:
            replica_selector = self.replica_selector

        if replica_selector is not None:
            dsn = replica_selector.current()
        else:
            dsn = self._dsn

        while True:
            try:
                # cx_Oracle deprecated `twophase` in 5.3 and removed
                # it in 6.0 in favor of the `internal_name` and
                # `external_name` parameters. In 5.3, these were both
                # implicitly set by `twophase` to "cx_Oracle".
                # According to the docs
                # (https://docs.oracle.com/cd/E18283_01/appdev.112/e10646/ociaahan.htm),
                # OCI_ATTR_EXTERNAL_NAME is "the user-friendly global
                # name stored in sys.props$.value$, where name =
                # 'GLOBAL_DB_NAME'. It is not guaranteed to be unique
                # " and OCI_ATTR_INTERNAL_NAME is "the client database
                # name that is recorded when performing global
                # transactions. The DBA can use the name to track
                # transactions"
                kw = {'threaded': True}
                conn = self._db_connect(self._user, self._password, dsn, **kw)
                conn.outputtypehandler = self._outputtypehandler
                if twophase:
                    conn.internal_name = 'cx_Oracle'
                    conn.external_name = 'cx_Oracle'
                cursor = conn.cursor()
                cursor.arraysize = 64

                if isolation:
                    cursor.execute("SET TRANSACTION %s" % isolation)
                return conn, cursor

            except self.driver.use_replica_exceptions as e:
                if replica_selector is not None:
                    next_dsn = next(replica_selector)
                    if next_dsn is not None:
                        log.warning("Unable to connect to DSN %s: %s, "
                                    "now trying %s", dsn, e, next_dsn)
                        dsn = next_dsn
                        continue
                log.warning("Unable to connect: %s", e)
                raise

    def _do_open_for_load(self):
        return self.open(self.isolation_read_only,
                         replica_selector=self.ro_replica_selector)

    def restart_load(self, conn, cursor, needs_rollback=True):
        """Reinitialize a connection for loading objects."""
        self.check_replica(conn, cursor,
                           replica_selector=self.ro_replica_selector)
        conn.rollback()
        cursor.execute("SET TRANSACTION %s" % self.isolation_read_only)

    def check_replica(self, conn, cursor, replica_selector=None):
        """Raise an exception if the connection belongs to an old replica"""
        if replica_selector is None:
            replica_selector = self.replica_selector

        if replica_selector is not None:
            current = replica_selector.current()
            if conn.dsn != current:
                # Prompt the change to a new replica by raising an exception.
                self.close(conn, cursor)
                raise ReplicaClosedException(
                    "Switched replica from %s to %s" % (conn.dsn, current))

    def _set_xid(self, conn, cursor):
        """Set up a distributed transaction"""
        stmt = """
        SELECT SYS_CONTEXT('USERENV', 'SID') FROM DUAL
        """
        cursor.execute(stmt)
        xid = str(cursor.fetchone()[0])
        conn.begin(0, xid, '0')

    def _do_open_for_store(self, **open_args):
        if self._twophase:
            conn, cursor = self.open(transaction_mode=None, twophase=True)
        else:
            conn, cursor = self.open()
        return conn, cursor

    def _after_opened_for_store(self, conn, cursor, restart=False):
        if self._twophase:
            self._set_xid(conn, cursor)

    def _outputtypehandler(self, cursor, name, defaultType,
                           size, precision, scale): # pylint:disable=unused-argument
        """
        cx_Oracle outputtypehandler that causes Oracle to send BLOBs
        inline. This works for sizes up to 1GB, which should be fine for object state,
        but not actual blobs.

        Note that if a BLOB in the result is too large, Oracle generates an
        error indicating truncation.  The run_lob_stmt() method works
        around this.
        """
        # pylint:disable=unused-argument
        if defaultType == self.driver.BLOB:
            # Default size for BLOB is 4000, we want the whole blob inline.
            return cursor.var(self.driver.LONG_BINARY, arraysize=cursor.arraysize)
        if defaultType == self.driver.CLOB:
            # Default size for CLOB is 4000, we want the whole blob inline.
            return cursor.var(self.driver.LONG_STRING, arraysize=cursor.arraysize)
